import numpy as np
import pytest
import xarray as xr

from aospy import Region
from aospy.region import (
    _get_land_mask,
    BoundsRect,
)
from aospy.internal_names import (
    LAT_STR,
    LON_STR,
    SFC_AREA_STR,
    LAND_MASK_STR
)
from aospy.utils import Longitude


@pytest.fixture()
def values_for_reg_arr():
    return np.array([[-2., 1.],
                     [np.nan, 5.],
                     [3., 3.],
                     [4., 4.2]])


@pytest.fixture()
def data_for_reg_calcs(values_for_reg_arr):
    lat = [-10., 1., 10., 20.]
    lon = [1., 10.]
    sfc_area = [0.5, 1., 0.5, 0.25]
    land_mask = [1., 1., 0., 1.]

    lat = xr.DataArray(lat, dims=[LAT_STR], coords=[lat])
    lon = xr.DataArray(lon, dims=[LON_STR], coords=[lon])
    sfc_area = xr.DataArray(sfc_area, dims=[LAT_STR], coords=[lat])
    land_mask = xr.DataArray(land_mask, dims=[LAT_STR], coords=[lat])

    sfc_area, _ = xr.broadcast(sfc_area, lon)
    land_mask, _ = xr.broadcast(land_mask, lon)

    da = xr.DataArray(values_for_reg_arr, coords=[lat, lon])
    da.coords[SFC_AREA_STR] = sfc_area
    da.coords[LAND_MASK_STR] = land_mask
    return da


_alt_names = {LON_STR: 'LONS', LAT_STR: 'LATS', LAND_MASK_STR: 'lm',
              SFC_AREA_STR: 'AREA'}


@pytest.fixture()
def data_reg_alt_names(data_for_reg_calcs):
    return data_for_reg_calcs.rename(_alt_names)


region_no_land_mask = Region(
    name='test',
    description='Test region with no land mask',
    west_bound=0.,
    east_bound=5,
    south_bound=0,
    north_bound=90.,
    do_land_mask=False
)


region_land_mask = Region(
    name='test',
    description='Test region with land mask',
    west_bound=0.,
    east_bound=5,
    south_bound=0,
    north_bound=90.,
    do_land_mask=True
)


_expected_mask = [[False, False],
                  [True, False],
                  [True, False],
                  [True, False]]


def test_get_land_mask_without_land_mask(data_for_reg_calcs):
    result = _get_land_mask(data_for_reg_calcs,
                            region_no_land_mask.do_land_mask)
    expected = 1
    assert result == expected


def test_get_land_mask_with_land_mask(data_for_reg_calcs):
    result = _get_land_mask(data_for_reg_calcs, region_land_mask.do_land_mask)
    expected = data_for_reg_calcs[LAND_MASK_STR]
    xr.testing.assert_identical(result, expected)


def test_get_land_mask_non_aospy_name(data_reg_alt_names):
    result = _get_land_mask(data_reg_alt_names, region_land_mask.do_land_mask,
                            land_mask_str=_alt_names[LAND_MASK_STR])
    expected = data_reg_alt_names[_alt_names[LAND_MASK_STR]]
    xr.testing.assert_identical(result, expected)


def test_region_init():
    region = Region(
        name='test',
        description='region description',
        west_bound=0.,
        east_bound=5,
        south_bound=0,
        north_bound=90.,
        do_land_mask=True
    )
    assert region.name == 'test'
    assert region.description == 'region description'
    assert isinstance(region.mask_bounds, tuple)
    assert len(region.mask_bounds) == 1
    assert isinstance(region.mask_bounds[0], BoundsRect)
    assert np.all(region.mask_bounds[0] ==
                  (Longitude(0.), Longitude(5), 0, 90.))
    assert region.do_land_mask is True


def test_region_init_mult_rect():
    bounds_in = [[1, 2, 3, 4], (-12, -30, 2.3, 9)]
    region = Region(name='test', mask_bounds=bounds_in)
    assert isinstance(region.mask_bounds, tuple)
    assert len(region.mask_bounds) == 2
    for (w, e, s, n), bounds in zip(bounds_in, region.mask_bounds):
        assert isinstance(bounds, tuple)
        assert np.all(bounds == (Longitude(w), Longitude(e), s, n))


def test_region_init_bad_bounds():
    with pytest.raises(ValueError):
        Region(mask_bounds=[(1, 2, 3)])
        Region(mask_bounds=[(1, 2, 3, 4),
                            (1, 2, 3)])


def test_make_mask_single_rect(data_for_reg_calcs):
    result = region_land_mask._make_mask(data_for_reg_calcs)
    expected = xr.DataArray(_expected_mask, dims=[LAT_STR, LON_STR],
                            coords={LAT_STR: data_for_reg_calcs[LAT_STR],
                                    LON_STR: data_for_reg_calcs[LON_STR]})
    xr.testing.assert_equal(result.transpose(), expected)


def test_make_mask_mult_rect(data_for_reg_calcs):
    mask_bounds = (region_land_mask.mask_bounds[0], [0, 360, -20, -5])
    region = Region(name='mult_rect', mask_bounds=mask_bounds)
    result = region._make_mask(data_for_reg_calcs)
    expected_values = [[True, True],
                       [True, False],
                       [True, False],
                       [True, False]]
    expected = xr.DataArray(expected_values, dims=[LAT_STR, LON_STR],
                            coords={LAT_STR: data_for_reg_calcs[LAT_STR],
                                    LON_STR: data_for_reg_calcs[LON_STR]})
    xr.testing.assert_equal(result.transpose(), expected)


@pytest.mark.parametrize(
    'region',
    [region_no_land_mask, region_land_mask],
    ids=['no-land-mask', 'land-mask'])
def test_mask_var(data_for_reg_calcs, region):
    # Test region masks first row and second column of test data.  Note that
    # first element of second row is np.nan in initial dataset.
    expected_data = [[np.nan, np.nan],
                     [np.nan, np.nan],
                     [3., np.nan],
                     [4., np.nan]]
    expected = data_for_reg_calcs.copy(deep=True)
    expected.values = expected_data
    result = region.mask_var(data_for_reg_calcs)
    xr.testing.assert_identical(result, expected)


@pytest.mark.parametrize(
    'region',
    [region_no_land_mask, region_land_mask],
    ids=['no-land-mask', 'land-mask'])
def test_mask_var_non_aospy_names(data_reg_alt_names, region):
    # Test region masks first row and second column of test data.  Note that
    # first element of second row is np.nan in initial dataset.
    expected_data = [[np.nan, np.nan],
                     [np.nan, np.nan],
                     [3., np.nan],
                     [4., np.nan]]
    expected = data_reg_alt_names.copy(deep=True)
    expected.values = expected_data
    result = region.mask_var(data_reg_alt_names, lon_str=_alt_names[LON_STR],
                             lat_str=_alt_names[LAT_STR])
    xr.testing.assert_identical(result, expected)


def test_ts_no_land_mask(data_for_reg_calcs):
    result = region_no_land_mask.ts(data_for_reg_calcs)

    data = data_for_reg_calcs.values
    sfc_area = data_for_reg_calcs.sfc_area.values
    exp_numerator = data[2, 0] * sfc_area[2, 0] + data[3, 0] * sfc_area[3, 0]
    exp_denominator = sfc_area[2, 0] + sfc_area[3, 0]
    expected = xr.DataArray(exp_numerator / exp_denominator)
    xr.testing.assert_identical(result, expected)


def test_ts_land_mask(data_for_reg_calcs):
    result = region_land_mask.ts(data_for_reg_calcs)
    expected = xr.DataArray(data_for_reg_calcs.values[3, 0])
    xr.testing.assert_identical(result, expected)


_map_to_alt_names = {'lon_str': _alt_names[LON_STR],
                     'lat_str': _alt_names[LAT_STR],
                     'land_mask_str': _alt_names[LAND_MASK_STR],
                     'sfc_area_str': _alt_names[SFC_AREA_STR]}


def test_ts_non_aospy_names(data_reg_alt_names):
    result = region_land_mask.ts(data_reg_alt_names, **_map_to_alt_names)
    expected = xr.DataArray(data_reg_alt_names.values[3, 0])
    xr.testing.assert_identical(result, expected)



"""Functionality for performing user-specified calculations on aospy data."""
from collections import OrderedDict
import logging
import os
import shutil
import subprocess
import tarfile
from time import ctime

import numpy as np
import xarray as xr

from ._constants import GRAV_EARTH
from .var import Var
from . import internal_names
from . import utils

logging.basicConfig(level=logging.INFO)


_P_VARS = {internal_names.ETA_STR: utils.vertcoord.p_eta,
           'pressure': utils.vertcoord.p_level}
_DP_VARS = {internal_names.ETA_STR: utils.vertcoord.dp_eta,
            'pressure': utils.vertcoord.dp_level}
_TIME_DEFINED_REDUCTIONS = ['av', 'std', 'ts', 'reg.av', 'reg.std', 'reg.ts']


def _replace_pressure(arguments, dtype_in_vert):
    """Replace p and dp Vars with appropriate Var objects specific to
    the dtype_in_vert."""
    arguments_out = []
    for arg in arguments:
        if isinstance(arg, Var):
            if arg.name == 'p':
                arguments_out.append(_P_VARS[dtype_in_vert])
            elif arg.name == 'dp':
                arguments_out.append(_DP_VARS[dtype_in_vert])
            else:
                arguments_out.append(arg)
        else:
            arguments_out.append(arg)
    return arguments_out


class Calc(object):
    """Class for executing, saving, and loading a single computation."""

    ARR_XARRAY_NAME = 'aospy_result'

    _grid_coords = [internal_names.LAT_STR, internal_names.LAT_BOUNDS_STR,
                    internal_names.LON_STR, internal_names.LON_BOUNDS_STR,
                    internal_names.ZSURF_STR, internal_names.SFC_AREA_STR,
                    internal_names.LAND_MASK_STR, internal_names.PK_STR,
                    internal_names.BK_STR, internal_names.PHALF_STR,
                    internal_names.PFULL_STR, internal_names.PLEVEL_STR]
    _grid_attrs = OrderedDict([(key, internal_names.GRID_ATTRS[key])
                               for key in _grid_coords])

    def __str__(self):
        """String representation of the object."""
        return "<aospy.Calc instance: " + ', '.join(
            (self.name, self.proj.name, self.model.name, self.run.name)
        ) + ">"

    __repr__ = __str__

    def _dir_out(self):
        """Create string of the data directory to save individual .nc files."""
        return os.path.join(self.proj.direc_out, self.proj.name,
                            self.model.name, self.run.name, self.name)

    def _dir_tar_out(self):
        """Create string of the data directory to store a tar file."""
        return os.path.join(self.proj.tar_direc_out, self.proj.name,
                            self.model.name, self.run.name)

    def _file_name(self, dtype_out_time, extension='nc'):
        """Create the name of the aospy file."""
        if dtype_out_time is None:
            dtype_out_time = ''
        out_lbl = utils.io.data_out_label(self.intvl_out, dtype_out_time,
                                          dtype_vert=self.dtype_out_vert)
        in_lbl = utils.io.data_in_label(self.intvl_in, self.dtype_in_time,
                                        self.dtype_in_vert)
        start_year = utils.times.infer_year(self.start_date)
        end_year = utils.times.infer_year(self.end_date)
        yr_lbl = utils.io.yr_label((start_year, end_year))
        return '.'.join(
            [self.name, out_lbl, in_lbl, self.model.name,
             self.run.name, yr_lbl, extension]
        ).replace('..', '.')

    def _path_out(self, dtype_out_time):
        return os.path.join(self.dir_out, self.file_name[dtype_out_time])

    def _path_tar_out(self):
        return os.path.join(self.dir_tar_out, 'data.tar')

    @staticmethod
    def _print_verbose(*args):
        """Print diagnostic message."""
        try:
            return '{0} {1} ({2})'.format(args[0], args[1], ctime())
        except IndexError:
            return '{0} ({1})'.format(args[0], ctime())

    def __init__(self, proj=None, model=None, run=None, var=None,
                 date_range=None, region=None, intvl_in=None, intvl_out=None,
                 dtype_in_time=None, dtype_in_vert=None, dtype_out_time=None,
                 dtype_out_vert=None, level=None, time_offset=None):
        """Instantiate a Calc object.

        Parameters
        ----------
        proj : aospy.Proj object
            The project for this calculation.
        model : aospy.Model object
            The model for this calculation.
        run : aospy.Run object
            The run for this calculation.
        var : aospy.Var object
            The variable for this calculation.
        region : sequence of aospy.Region objects
            The region(s) over which any regional reductions will be performed.
        date_range : tuple of datetime.datetime objects
            The range of dates over which to perform calculations.
        intvl_in : {None, 'annual', 'monthly', 'daily', '6hr', '3hr'}, optional
            The time resolution of the input data.
        dtype_in_time : {None, 'inst', 'ts', 'av', 'av_ts'}, optional
            What the time axis of the input data represents:

            - 'inst' : Timeseries of instantaneous values
            - 'ts' : Timeseries of averages over the period of each time-index
            - 'av' : A single value averaged over a date range

        dtype_in_vert : {None, 'pressure', 'sigma'}, optional
            The vertical coordinate system used by the input data:

            - None : not defined vertically
            - 'pressure' : pressure coordinates
            - 'sigma' : hybrid sigma-pressure coordinates

        intvl_out : {'ann', season-string, month-integer}
            The sub-annual time interval over which to compute:

            - 'ann' : Annual mean
            - season-string : E.g. 'JJA' for June-July-August
            - month-integer : 1 for January, 2 for February, etc.

        dtype_out_time : tuple with elements being one or more of:

            - Gridpoint-by-gridpoint output:

              - 'av' : Gridpoint-by-gridpoint time-average
              - 'std' : Gridpoint-by-gridpoint temporal standard deviation
              - 'ts' : Gridpoint-by-gridpoint time-series

            - Averages over each region specified via `region`:

              - 'reg.av', 'reg.std', 'reg.ts' : analogous to 'av', 'std', 'ts'

        dtype_out_vert : {None, 'vert_av', 'vert_int'}, optional
            How to reduce the data vertically:

            - None : no vertical reduction (i.e. output is defined vertically)
            - 'vert_av' : mass-weighted vertical average
            - 'vert_int' : mass-weighted vertical integral

        time_offset : {None, dict}, optional
            How to offset input data in time to correct for metadata errors

            - None : no time offset applied
            - dict : e.g. ``{'hours': -3}`` to offset times by -3 hours
              See :py:meth:`aospy.utils.times.apply_time_offset`.

        """
        if run not in model.runs:
            raise AttributeError("Model '{0}' has no run '{1}'.  Calc object "
                                 "will not be generated.".format(model, run))
        self.proj = proj
        self.model = model

        self.run = run
        self.default_start_date = self.run.default_start_date
        self.default_end_date = self.run.default_end_date
        self.data_loader = self.run.data_loader

        self.var = var
        self.name = self.var.name
        self.domain = self.var.domain
        self.def_time = self.var.def_time
        self.def_vert = self.var.def_vert

        logging.debug(self._print_verbose('Initializing Calc '
                                          'instance:', self.__str__()))

        try:
            self.function = self.var.func
        except AttributeError:
            self.function = lambda x: x
        if getattr(self.var, 'variables', False):
            self.variables = self.var.variables
        else:
            self.variables = (self.var,)

        self.intvl_in = intvl_in
        self.intvl_out = intvl_out
        self.dtype_in_time = dtype_in_time
        self.dtype_in_vert = dtype_in_vert

        if isinstance(dtype_out_time, (list, tuple)):
            self.dtype_out_time = tuple(dtype_out_time)
        else:
            self.dtype_out_time = tuple([dtype_out_time])
        if not self.def_time:
            for reduction in self.dtype_out_time:
                if reduction in _TIME_DEFINED_REDUCTIONS:
                    msg = ("Var {0} has no time dimension "
                           "for the given time reduction "
                           "{1}".format(self.name, reduction))
                    raise ValueError(msg)

        self.level = level
        self.dtype_out_vert = dtype_out_vert
        self.region = region

        self.months = utils.times.month_indices(intvl_out)
        if date_range == 'default':
            self.start_date = utils.times.ensure_datetime(
                self.run.default_start_date)
            self.end_date = utils.times.ensure_datetime(
                self.run.default_end_date)
        else:
            self.start_date = utils.times.ensure_datetime(date_range[0])
            self.end_date = utils.times.ensure_datetime(date_range[-1])

        self.time_offset = time_offset
        self.data_loader_attrs = dict(
            domain=self.domain, intvl_in=self.intvl_in,
            dtype_in_vert=self.dtype_in_vert,
            dtype_in_time=self.dtype_in_time, intvl_out=self.intvl_out)

        self.model.set_grid_data()

        self.dir_out = self._dir_out()
        self.dir_tar_out = self._dir_tar_out()
        self.file_name = {d: self._file_name(d) for d in self.dtype_out_time}
        self.path_out = {d: self._path_out(d)
                         for d in self.dtype_out_time}
        self.path_tar_out = self._path_tar_out()

        self.data_out = {}

    def _to_desired_dates(self, arr):
        """Restrict the xarray DataArray or Dataset to the desired months."""
        times = utils.times.extract_months(
            arr[internal_names.TIME_STR], self.months
        )
        return arr.sel(time=times)

    def _add_grid_attributes(self, ds):
        """Add model grid attributes to a dataset"""
        for name_int, names_ext in self._grid_attrs.items():
            ds_coord_name = set(names_ext).intersection(set(ds.coords) |
                                                        set(ds.data_vars))
            model_attr = getattr(self.model, name_int, None)
            if ds_coord_name and (model_attr is not None):
                # Force coords to have desired name.
                ds = ds.rename({list(ds_coord_name)[0]: name_int})
                ds = ds.set_coords(name_int)
                if not np.array_equal(ds[name_int], model_attr):
                    if np.allclose(ds[name_int], model_attr):
                        msg = ("Values for '{0}' are nearly (but not exactly) "
                               "the same in the Run {1} and the Model {2}.  "
                               "Therefore replacing Run's values with the "
                               "model's.".format(name_int, self.run,
                                                 self.model))
                        logging.info(msg)
                        ds[name_int].values = model_attr.values
                    else:
                        msg = ("Model coordinates for '{0}' do not match those"
                               " in Run: {1} vs. {2}"
                               "".format(name_int, ds[name_int], model_attr))
                        logging.info(msg)

            else:
                # Bring in coord from model object if it exists.
                ds = ds.load()
                if model_attr is not None:
                    ds[name_int] = model_attr
                    ds = ds.set_coords(name_int)
            if (self.dtype_in_vert == 'pressure' and
                    internal_names.PLEVEL_STR in ds.coords):
                self.pressure = ds.level
        return ds

    def _get_input_data(self, var, start_date, end_date):
        """Get the data for a single variable over the desired date range."""
        logging.info(self._print_verbose("Getting input data:", var))

        if isinstance(var, (float, int)):
            return var
        else:
            cond_pfull = ((not hasattr(self, internal_names.PFULL_STR))
                          and var.def_vert and
                          self.dtype_in_vert == internal_names.ETA_STR)
            data = self.data_loader.recursively_compute_variable(
                var, start_date, end_date, self.time_offset, self.model,
                **self.data_loader_attrs)
            name = data.name
            data = self._add_grid_attributes(data.to_dataset(name=data.name))
            data = data[name]
            if cond_pfull:
                try:
                    self.pfull_coord = data[internal_names.PFULL_STR]
                except KeyError:
                    pass
            # Force all data to be at full pressure levels, not half levels.
            bool_to_pfull = (self.dtype_in_vert == internal_names.ETA_STR and
                             var.def_vert == internal_names.PHALF_STR)
            if bool_to_pfull:
                data = utils.vertcoord.to_pfull_from_phalf(data,
                                                           self.pfull_coord)
        if var.def_time:
            # Restrict to the desired dates within each year.
            if self.dtype_in_time != 'av':
                return self._to_desired_dates(data)
        else:
            return data

    def _get_all_data(self, start_date, end_date):
        """Get the needed data from all of the vars in the calculation."""
        return [self._get_input_data(var, start_date, end_date)
                for var in _replace_pressure(self.variables,
                                             self.dtype_in_vert)]

    def _local_ts(self, *data):
        """Perform the computation at each gridpoint and time index."""
        return self.function(*data).rename(self.name)

    def _compute(self, data):
        """Perform the calculation."""
        local_ts = self._local_ts(*data)
        dt = local_ts[internal_names.TIME_WEIGHTS_STR]
        # Convert dt to units of days to prevent overflow
        dt = dt / np.timedelta64(1, 'D')
        return local_ts, dt

    def _compute_full_ts(self, data):
        """Perform calculation and create yearly timeseries at each point."""
        # Get results at each desired timestep and spatial point.
        full_ts, dt = self._compute(data)
        # Vertically integrate.
        vert_types = ('vert_int', 'vert_av')
        if self.dtype_out_vert in vert_types and self.var.def_vert:
            dp = self._get_input_data(_DP_VARS[self.dtype_in_vert],
                                      self.start_date, self.end_date)
            full_ts = utils.vertcoord.int_dp_g(full_ts, dp)
            if self.dtype_out_vert == 'vert_av':
                ps = self._get_input_data(utils.vertcoord.ps,
                                          self.start_date, self.end_date)
                full_ts *= (GRAV_EARTH / ps)
        return full_ts, dt

    def _full_to_yearly_ts(self, arr, dt):
        """Average the full timeseries within each year."""
        time_defined = self.def_time and not ('av' in self.dtype_in_time)
        if time_defined:
            arr = utils.times.yearly_average(arr, dt)
        return arr

    def _time_reduce(self, arr, reduction):
        """Perform the specified time reduction on a local time-series."""
        if self.dtype_in_time == 'av' or not self.def_time:
            return arr
        reductions = {
            'ts': lambda xarr: xarr,
            'av': lambda xarr: xarr.mean(internal_names.YEAR_STR),
            'std': lambda xarr: xarr.std(internal_names.YEAR_STR),
            }
        try:
            return reductions[reduction](arr)
        except KeyError:
            raise ValueError("Specified time-reduction method '{}' is not "
                             "supported".format(reduction))

    def region_calcs(self, arr, func):
        """Perform a calculation for all regions."""
        # Get pressure values for data output on hybrid vertical coordinates.
        bool_pfull = (self.def_vert and self.dtype_in_vert ==
                      internal_names.ETA_STR and self.dtype_out_vert is False)
        if bool_pfull:
            pfull_data = self._get_input_data(_P_VARS[self.dtype_in_vert],
                                              self.start_date,
                                              self.end_date)
            pfull = self._full_to_yearly_ts(
                pfull_data, arr[internal_names.TIME_WEIGHTS_STR]
            ).rename('pressure')
        # Loop over the regions, performing the calculation.
        reg_dat = {}
        for reg in self.region:
            # Just pass along the data if averaged already.
            if 'av' in self.dtype_in_time:
                data_out = reg.ts(arr)
            # Otherwise perform the calculation.
            else:
                method = getattr(reg, func)
                data_out = method(arr)
                if bool_pfull:
                    # Don't apply e.g. standard deviation to coordinates.
                    if func not in ['av', 'ts']:
                        method = reg.ts
                    # Convert Pa to hPa
                    coord = method(pfull) * 1e-2
                    data_out = data_out.assign_coords(
                        **{reg.name + '_pressure': coord}
                    )
            reg_dat.update(**{reg.name: data_out})
        return xr.Dataset(reg_dat)

    def _apply_all_time_reductions(self, data):
        """Apply all requested time reductions to the data."""
        logging.info(self._print_verbose("Applying desired time-"
                                         "reduction methods."))
        reduc_specs = [r.split('.') for r in self.dtype_out_time]
        reduced = {}
        for reduc, specs in zip(self.dtype_out_time, reduc_specs):
            func = specs[-1]
            if 'reg' in specs:
                reduced.update({reduc: self.region_calcs(data, func)})
            else:
                reduced.update({reduc: self._time_reduce(data, func)})
        return OrderedDict(sorted(reduced.items(), key=lambda t: t[0]))

    def compute(self, write_to_tar=True):
        """Perform all desired calculations on the data and save externally."""
        data = self._get_all_data(self.start_date, self.end_date)
        logging.info('Computing timeseries for {0} -- '
                     '{1}.'.format(self.start_date, self.end_date))
        full, full_dt = self._compute_full_ts(data)
        full_out = self._full_to_yearly_ts(full, full_dt)
        reduced = self._apply_all_time_reductions(full_out)
        logging.info("Writing desired gridded outputs to disk.")
        for dtype_time, data in reduced.items():
            data = _add_metadata_as_attrs(data, self.var.units,
                                          self.var.description,
                                          self.dtype_out_vert)
            self.save(data, dtype_time, dtype_out_vert=self.dtype_out_vert,
                      save_files=True, write_to_tar=write_to_tar)
        return self

    def _save_files(self, data, dtype_out_time):
        """Save the data to netcdf files in direc_out."""
        path = self.path_out[dtype_out_time]
        if not os.path.isdir(self.dir_out):
            os.makedirs(self.dir_out)
        if 'reg' in dtype_out_time:
            try:
                reg_data = xr.open_dataset(path)
            except (EOFError, RuntimeError, IOError):
                reg_data = xr.Dataset()
            reg_data.update(data)
            data_out = reg_data
        else:
            data_out = data
        if isinstance(data_out, xr.DataArray):
            data_out = xr.Dataset({self.name: data_out})
        data_out.to_netcdf(path, engine='netcdf4', format='NETCDF3_64BIT')

    def _write_to_tar(self, dtype_out_time):
        """Add the data to the tar file in tar_out_direc."""
        # When submitted in parallel and the directory does not exist yet
        # multiple processes may try to create a new directory; this leads
        # to an OSError for all processes that tried to make the
        # directory, but were later than the first.
        try:
            os.makedirs(self.dir_tar_out)
        except OSError:
            pass
        # tarfile 'append' mode won't overwrite the old file, which we want.
        # So open in 'read' mode, extract the file, and then delete it.
        # But 'read' mode throws OSError if file doesn't exist: make it first.
        utils.io.dmget([self.path_tar_out])
        with tarfile.open(self.path_tar_out, 'a') as tar:
            pass
        with tarfile.open(self.path_tar_out, 'r') as tar:
            old_data_path = os.path.join(self.dir_tar_out,
                                         self.file_name[dtype_out_time])
            try:
                tar.extract(self.file_name[dtype_out_time],
                            path=old_data_path)
            except KeyError:
                pass
            else:
                # The os module treats files on archive as non-empty
                # directories, so can't use os.remove or os.rmdir.
                shutil.rmtree(old_data_path)
                retcode = subprocess.call([
                    "tar", "--delete", "--file={}".format(self.path_tar_out),
                    self.file_name[dtype_out_time]
                ])
                if retcode:
                    msg = ("The 'tar' command to save your aospy output "
                           "exited with an error.  Most likely, this is due "
                           "to using an old version of 'tar' (especially if "
                           "you are on a Mac).  Consider installing a newer "
                           "version of 'tar' or disabling tar output by "
                           "setting `write_to_tar=False` in the "
                           "`calc_exec_options` argument of "
                           "`submit_mult_calcs`.")
                    logging.warn(msg)
        with tarfile.open(self.path_tar_out, 'a') as tar:
            tar.add(self.path_out[dtype_out_time],
                    arcname=self.file_name[dtype_out_time])

    def _update_data_out(self, data, dtype):
        """Append the data of the given dtype_out to the data_out attr."""
        try:
            self.data_out.update({dtype: data})
        except AttributeError:
            self.data_out = {dtype: data}

    def save(self, data, dtype_out_time, dtype_out_vert=False,
             save_files=True, write_to_tar=False):
        """Save aospy data to data_out attr and to an external file."""
        self._update_data_out(data, dtype_out_time)
        if save_files:
            self._save_files(data, dtype_out_time)
        if write_to_tar and self.proj.tar_direc_out:
            self._write_to_tar(dtype_out_time)
        logging.info('\t{}'.format(self.path_out[dtype_out_time]))

    def _load_from_disk(self, dtype_out_time, dtype_out_vert=False,
                        region=False):
        """Load aospy data saved as netcdf files on the file system."""
        ds = xr.open_dataset(self.path_out[dtype_out_time])
        if region:
            arr = ds[region.name]
            # Use region-specific pressure values if available.
            if (self.dtype_in_vert == internal_names.ETA_STR
                    and not dtype_out_vert):
                reg_pfull_str = region.name + '_pressure'
                arr = arr.drop([r for r in arr.coords.iterkeys()
                                if r not in (internal_names.PFULL_STR,
                                             reg_pfull_str)])
                # Rename pfull to pfull_ref always.
                arr = arr.rename({internal_names.PFULL_STR:
                                  internal_names.PFULL_STR + '_ref'})
                # Rename region_pfull to pfull if its there.
                if hasattr(arr, reg_pfull_str):
                    return arr.rename({reg_pfull_str:
                                       internal_names.PFULL_STR})
                return arr
            return arr
        return ds[self.name]

    def _load_from_tar(self, dtype_out_time, dtype_out_vert=False):
        """Load data save in tarball form on the file system."""
        path = os.path.join(self.dir_tar_out, 'data.tar')
        utils.io.dmget([path])
        with tarfile.open(path, 'r') as data_tar:
            ds = xr.open_dataset(
                data_tar.extractfile(self.file_name[dtype_out_time])
            )
            return ds[self.name]

    def load(self, dtype_out_time, dtype_out_vert=False, region=False,
             plot_units=False, mask_unphysical=False):
        """Load the data from the object if possible or from disk."""
        msg = ("Loading data from disk for object={0}, dtype_out_time={1}, "
               "dtype_out_vert={2}, and region="
               "{3}".format(self, dtype_out_time, dtype_out_vert, region))
        logging.info(msg + ' ({})'.format(ctime()))
        # Grab from the object if its there.
        try:
            data = self.data_out[dtype_out_time]
        except (AttributeError, KeyError):
            # Otherwise get from disk.  Try scratch first, then archive.
            try:
                data = self._load_from_disk(dtype_out_time, dtype_out_vert,
                                            region=region)
            except IOError:
                data = self._load_from_tar(dtype_out_time, dtype_out_vert)
        # Copy the array to self.data_out for ease of future access.
        self._update_data_out(data, dtype_out_time)
        # Apply desired plotting/cleanup methods.
        if mask_unphysical:
            data = self.var.mask_unphysical(data)
        if plot_units:
            data = self.var.to_plot_units(data, dtype_vert=dtype_out_vert)
        return data


def _add_metadata_as_attrs(data, units, description, dtype_out_vert):
    """Add metadata attributes to Dataset or DataArray"""
    if isinstance(data, xr.DataArray):
        return _add_metadata_as_attrs_da(data, units, description,
                                         dtype_out_vert)
    else:
        for name, arr in data.data_vars.items():
            _add_metadata_as_attrs_da(arr, units, description,
                                      dtype_out_vert)
        return data


def _add_metadata_as_attrs_da(data, units, description, dtype_out_vert):
    """Add metadata attributes to DataArray"""
    if dtype_out_vert == 'vert_int':
        if units != '':
            units = '(vertical integral of {0}): {0} kg m^-2)'.format(units)
        else:
            units = '(vertical integral of quantity with unspecified units)'
    data.attrs['units'] = units
    data.attrs['description'] = description
    return data



"""aospy DataLoader objects"""
import logging
import os
import pprint
import warnings

import numpy as np
import xarray as xr

from .internal_names import (
    ETA_STR,
    GRID_ATTRS,
    TIME_STR,
    TIME_BOUNDS_STR,
)
from .utils import times, io


def _preprocess_and_rename_grid_attrs(func, grid_attrs=None, **kwargs):
    """Call a custom preprocessing method first then rename grid attrs.

    This wrapper is needed to generate a single function to pass to the
    ``preprocesss`` of xr.open_mfdataset.  It makes sure that the
    user-specified preprocess function is called on the loaded Dataset before
    aospy's is applied.  An example for why this might be needed is output from
    the WRF model; one needs to add a CF-compliant units attribute to the time
    coordinate of all input files, because it is not present by default.

    Parameters
    ----------
    func : function
       An arbitrary function to call before calling
       ``grid_attrs_to_aospy_names`` in ``_load_data_from_disk``.  Must take
       an xr.Dataset as an argument as well as ``**kwargs``.
    grid_attrs : dict (optional)
        Overriding dictionary of grid attributes mapping aospy internal
        names to names of grid attributes used in a particular model.

    Returns
    -------
    function
        A function that calls the provided function ``func`` on the Dataset
        before calling ``grid_attrs_to_aospy_names``; this is meant to be
        passed as a ``preprocess`` argument to ``xr.open_mfdataset``.
    """

    def func_wrapper(ds):
        return grid_attrs_to_aospy_names(func(ds, **kwargs), grid_attrs)
    return func_wrapper


def grid_attrs_to_aospy_names(data, grid_attrs=None):
    """Rename grid attributes to be consistent with aospy conventions.

    Search all of the dataset's coords and dims looking for matches to known
    grid attribute names; any that are found subsequently get renamed to the
    aospy name as specified in ``aospy.internal_names.GRID_ATTRS``.

    Also forces any renamed grid attribute that is saved as a dim without a
    coord to have a coord, which facilitates subsequent slicing/subsetting.

    This function does not compare to Model coordinates or add missing
    coordinates from Model objects.

    Parameters
    ----------
    data : xr.Dataset
    grid_attrs : dict (default None)
        Overriding dictionary of grid attributes mapping aospy internal
        names to names of grid attributes used in a particular model.

    Returns
    -------
    xr.Dataset
        Data returned with coordinates consistent with aospy
        conventions
    """
    if grid_attrs is None:
        grid_attrs = {}

    # Override GRID_ATTRS with entries in grid_attrs
    attrs = GRID_ATTRS.copy()
    for k, v in grid_attrs.items():
        if k not in attrs:
            raise ValueError(
                'Unrecognized internal name, {!r}, specified for a custom '
                'grid attribute name.  See the full list of valid internal '
                'names below:\n\n{}'.format(k, list(GRID_ATTRS.keys())))
        attrs[k] = (v, )

    dims_and_vars = set(data.variables).union(set(data.dims))
    for name_int, names_ext in attrs.items():
        data_coord_name = set(names_ext).intersection(dims_and_vars)
        if data_coord_name:
            data = data.rename({data_coord_name.pop(): name_int})
    return set_grid_attrs_as_coords(data)


def set_grid_attrs_as_coords(ds):
    """Set available grid attributes as coordinates in a given Dataset.

    Grid attributes are assumed to have their internal aospy names. Grid
    attributes are set as coordinates, such that they are carried by all
    selected DataArrays with overlapping index dimensions.

    Parameters
    ----------
    ds : Dataset
        Input data

    Returns
    -------
    Dataset
        Dataset with grid attributes set as coordinates
    """
    grid_attrs_in_ds = set(GRID_ATTRS.keys()).intersection(
        set(ds.coords) | set(ds.data_vars))
    ds = ds.set_coords(grid_attrs_in_ds)
    return ds


def _maybe_cast_to_float64(da):
    """Cast DataArrays to np.float64 if they are of type np.float32.

    Parameters
    ----------
    da : xr.DataArray
        Input DataArray

    Returns
    -------
    DataArray
    """
    if da.dtype == np.float32:
        logging.warning('Datapoints were stored using the np.float32 datatype.'
                        'For accurate reduction operations using bottleneck, '
                        'datapoints are being cast to the np.float64 datatype.'
                        ' For more information see: https://github.com/pydata/'
                        'xarray/issues/1346')
        return da.astype(np.float64)
    else:
        return da


def _sel_var(ds, var, upcast_float32=True):
    """Select the specified variable by trying all possible alternative names.

    Parameters
    ----------
    ds : Dataset
        Dataset possibly containing var
    var : aospy.Var
        Variable to find data for
    upcast_float32 : bool (default True)
        Whether to cast a float32 DataArray up to float64

    Returns
    -------
    DataArray

    Raises
    ------
    KeyError
        If the variable is not in the Dataset
    """
    for name in var.names:
        try:
            da = ds[name].rename(var.name)
            if upcast_float32:
                return _maybe_cast_to_float64(da)
            else:
                return da
        except KeyError:
            pass
    msg = '{0} not found among names: {1} in\n{2}'.format(var, var.names, ds)
    raise LookupError(msg)


def _prep_time_data(ds):
    """Prepare time coordinate information in Dataset for use in aospy.

    1. If the Dataset contains a time bounds coordinate, add attributes
       representing the true beginning and end dates of the time interval used
       to construct the Dataset
    2. If the Dataset contains a time bounds coordinate, overwrite the time
       coordinate values with the averages of the time bounds at each timestep
    3. Decode the times into np.datetime64 objects for time indexing

    Parameters
    ----------
    ds : Dataset
        Pre-processed Dataset with time coordinate renamed to
        internal_names.TIME_STR

    Returns
    -------
    Dataset
        The processed Dataset
    """
    ds = times.ensure_time_as_index(ds)
    if TIME_BOUNDS_STR in ds:
        ds = times.ensure_time_avg_has_cf_metadata(ds)
        ds[TIME_STR] = times.average_time_bounds(ds)
    else:
        logging.warning("dt array not found.  Assuming equally spaced "
                        "values in time, even though this may not be "
                        "the case")
        ds = times.add_uniform_time_weights(ds)
    # Suppress enable_cftimeindex is a no-op warning; we'll keep setting it for
    # now to maintain backwards compatibility for older xarray versions.
    with warnings.catch_warnings():
        warnings.filterwarnings('ignore')
        with xr.set_options(enable_cftimeindex=True):
            ds = xr.decode_cf(ds, decode_times=True, decode_coords=False,
                              mask_and_scale=True)
    return ds


def _load_data_from_disk(file_set, preprocess_func=lambda ds: ds,
                         data_vars='minimal', coords='minimal',
                         grid_attrs=None, **kwargs):
    """Load a Dataset from a list or glob-string of files.

    Datasets from files are concatenated along time,
    and all grid attributes are renamed to their aospy internal names.

    Parameters
    ----------
    file_set : list or str
        List of paths to files or glob-string
    preprocess_func : function (optional)
        Custom function to call before applying any aospy logic
        to the loaded dataset
    data_vars : str (default 'minimal')
        Mode for concatenating data variables in call to ``xr.open_mfdataset``
    coords : str (default 'minimal')
        Mode for concatenating coordinate variables in call to
        ``xr.open_mfdataset``.
    grid_attrs : dict
        Overriding dictionary of grid attributes mapping aospy internal
        names to names of grid attributes used in a particular model.

    Returns
    -------
    Dataset
    """
    apply_preload_user_commands(file_set)
    func = _preprocess_and_rename_grid_attrs(preprocess_func, grid_attrs,
                                             **kwargs)
    return xr.open_mfdataset(file_set, preprocess=func, concat_dim=TIME_STR,
                             decode_times=False, decode_coords=False,
                             mask_and_scale=True, data_vars=data_vars,
                             coords=coords)


def apply_preload_user_commands(file_set, cmd=io.dmget):
    """Call desired functions on file list before loading.

    For example, on the NOAA Geophysical Fluid Dynamics Laboratory
    computational cluster, data that is saved on their tape archive
    must be accessed via a `dmget` (or `hsmget`) command before being used.
    """
    if cmd is not None:
        cmd(file_set)


def _setattr_default(obj, attr, value, default):
    """Set an attribute of an object to a value or default value."""
    if value is None:
        setattr(obj, attr, default)
    else:
        setattr(obj, attr, value)


class DataLoader(object):
    """A fundamental DataLoader object."""
    def load_variable(self, var=None, start_date=None, end_date=None,
                      time_offset=None, grid_attrs=None, **DataAttrs):
        """Load a DataArray for requested variable and time range.

        Automatically renames all grid attributes to match aospy conventions.

        Parameters
        ----------
        var : Var
            aospy Var object
        start_date : datetime.datetime
            start date for interval
        end_date : datetime.datetime
            end date for interval
        time_offset : dict
            Option to add a time offset to the time coordinate to correct for
            incorrect metadata.
        grid_attrs : dict (optional)
            Overriding dictionary of grid attributes mapping aospy internal
            names to names of grid attributes used in a particular model.
        **DataAttrs
            Attributes needed to identify a unique set of files to load from

        Returns
        -------
        da : DataArray
             DataArray for the specified variable, date range, and interval in
        """
        file_set = self._generate_file_set(var=var, start_date=start_date,
                                           end_date=end_date, **DataAttrs)
        ds = _load_data_from_disk(
            file_set, self.preprocess_func, data_vars=self.data_vars,
            coords=self.coords, start_date=start_date, end_date=end_date,
            time_offset=time_offset, grid_attrs=grid_attrs, **DataAttrs
        )
        if var.def_time:
            ds = _prep_time_data(ds)
            start_date = times.maybe_convert_to_index_date_type(
                ds.indexes[TIME_STR], start_date)
            end_date = times.maybe_convert_to_index_date_type(
                ds.indexes[TIME_STR], end_date)
        ds = set_grid_attrs_as_coords(ds)
        da = _sel_var(ds, var, self.upcast_float32)
        if var.def_time:
            da = self._maybe_apply_time_shift(da, time_offset, **DataAttrs)
            return times.sel_time(da, start_date, end_date).load()
        else:
            return da.load()

    def _load_or_get_from_model(self, var, start_date=None, end_date=None,
                                time_offset=None, model=None, **DataAttrs):
        """Load a DataArray for the requested variable and time range

        Supports both access of grid attributes either through the DataLoader
        or through an optionally-provided Model object.  Defaults to using
        the version found in the DataLoader first.
        """
        grid_attrs = None if model is None else model.grid_attrs

        try:
            return self.load_variable(
                var, start_date=start_date, end_date=end_date,
                time_offset=time_offset, grid_attrs=grid_attrs, **DataAttrs)
        except (KeyError, IOError) as e:
            if var.name not in GRID_ATTRS or model is None:
                raise e
            else:
                try:
                    return getattr(model, var.name)
                except AttributeError:
                    raise AttributeError(
                        'Grid attribute {} could not be located either '
                        'through this DataLoader or in the provided Model '
                        'object: {}.'.format(var, model))

    def recursively_compute_variable(self, var, start_date=None, end_date=None,
                                     time_offset=None, model=None,
                                     **DataAttrs):
        """Compute a variable recursively, loading data where needed.

        An obvious requirement here is that the variable must eventually be
        able to be expressed in terms of model-native quantities; otherwise the
        recursion will never stop.

        Parameters
        ----------
        var : Var
            aospy Var object
        start_date : datetime.datetime
            start date for interval
        end_date : datetime.datetime
            end date for interval
        time_offset : dict
            Option to add a time offset to the time coordinate to correct for
            incorrect metadata.
        model : Model
            aospy Model object (optional)
        **DataAttrs
            Attributes needed to identify a unique set of files to load from

        Returns
        -------
        da : DataArray
             DataArray for the specified variable, date range, and interval in
        """
        if var.variables is None:
            return self._load_or_get_from_model(
                var, start_date, end_date, time_offset, model, **DataAttrs)
        else:
            data = [self.recursively_compute_variable(
                v, start_date, end_date, time_offset, model, **DataAttrs)
                    for v in var.variables]
            return var.func(*data).rename(var.name)

    @staticmethod
    def _maybe_apply_time_shift(da, time_offset=None, **DataAttrs):
        """Apply specified time shift to DataArray"""
        if time_offset is not None:
            time = times.apply_time_offset(da[TIME_STR], **time_offset)
            da[TIME_STR] = time
        return da

    def _generate_file_set(self, var=None, start_date=None, end_date=None,
                           domain=None, intvl_in=None, dtype_in_vert=None,
                           dtype_in_time=None, intvl_out=None):
        raise NotImplementedError(
            'All DataLoaders require a _generate_file_set method')


class DictDataLoader(DataLoader):
    """A DataLoader that uses a dict mapping lists of files to string tags.

    This is the simplest DataLoader; it is useful for instance if one is
    dealing with raw model history files, which tend to group all variables
    of a single output interval into single filesets. The
    intvl_in parameter is a string description of the time frequency of the
    data one is referencing (e.g. 'monthly', 'daily', '3-hourly').  In
    principle, one can give it any string value.

    Parameters
    ----------
    file_map : dict
        A dict mapping an input interval to a list of files
    upcast_float32 : bool (default True)
        Whether to cast loaded DataArrays with the float32 datatype to float64
        before doing calculations
    data_vars : str (default 'minimal')
        Mode for concatenating data variables in call to ``xr.open_mfdataset``
    coords : str (default 'minimal')
        Mode for concatenating coordinate variables in call to
        ``xr.open_mfdataset``.
    preprocess_func : function (optional)
        A function to apply to every Dataset before processing in aospy.  Must
        take a Dataset and ``**kwargs`` as its two arguments.

    Examples
    --------
    Case of two sets of files, one with monthly average output, and one with
    3-hourly output.

    >>> file_map = {'monthly': '000[4-6]0101.atmos_month.nc',
    ...             '3hr': '000[4-6]0101.atmos_8xday.nc'}
    >>> data_loader = DictDataLoader(file_map)

    If one wanted to correct a CF-incompliant units attribute on each Dataset
    read in, which depended on the ``intvl_in`` of the fileset one could
    define a ``preprocess_func`` which took into account the ``intvl_in``
    keyword argument.

    >>> def preprocess(ds, **kwargs):
    ...     if kwargs['intvl_in'] == 'monthly':
    ...         ds['time'].attrs['units'] = 'days since 0001-01-0000'
    ...     if kwargs['intvl_in'] == '3hr':
    ...         ds['time'].attrs['units'] = 'hours since 0001-01-0000'
    ...     return ds
    >>> data_loader = DictDataLoader(file_map, preprocess)
    """
    def __init__(self, file_map=None, upcast_float32=True, data_vars='minimal',
                 coords='minimal', preprocess_func=lambda ds, **kwargs: ds):
        """Create a new DictDataLoader."""
        self.file_map = file_map
        self.upcast_float32 = upcast_float32
        self.data_vars = data_vars
        self.coords = coords
        self.preprocess_func = preprocess_func

    def _generate_file_set(self, var=None, start_date=None, end_date=None,
                           domain=None, intvl_in=None, dtype_in_vert=None,
                           dtype_in_time=None, intvl_out=None):
        """Returns the file_set for the given interval in."""
        try:
            return self.file_map[intvl_in]
        except KeyError:
            raise KeyError('File set does not exist for the specified'
                           ' intvl_in {0}'.format(intvl_in))


class NestedDictDataLoader(DataLoader):
    """DataLoader that uses a nested dictionary mapping to load files.

    This is the most flexible existing type of DataLoader; it allows for the
    specification of different sets of files for different variables.  The
    intvl_in parameter is a string description of the time frequency of the
    data one is referencing (e.g. 'monthly', 'daily', '3-hourly').  In
    principle, one can give it any string value.  The variable name
    can be any variable name in your aospy object library (including
    alternative names).

    Parameters
    ----------
    file_map : dict
        A dict mapping intvl_in to dictionaries mapping Var
        objects to lists of files
    upcast_float32 : bool (default True)
        Whether to cast loaded DataArrays with the float32 datatype to float64
        before doing calculations
    data_vars : str (default 'minimal')
        Mode for concatenating data variables in call to ``xr.open_mfdataset``
    coords : str (default 'minimal')
        Mode for concatenating coordinate variables in call to
        ``xr.open_mfdataset``.
    preprocess_func : function (optional)
        A function to apply to every Dataset before processing in aospy.  Must
        take a Dataset and ``**kwargs`` as its two arguments.

    Examples
    --------
    Case of a set of monthly average files for large scale precipitation,
    and another monthly average set of files for convective precipitation.

    >>> file_map = {'monthly': {'precl': '000[4-6]0101.precl.nc',
    ...                         'precc': '000[4-6]0101.precc.nc'}}
    >>> data_loader = NestedDictDataLoader(file_map)

    See :py:class:`aospy.data_loader.DictDataLoader` for an example of a
    possible function to pass as a ``preprocess_func``.
    """
    def __init__(self, file_map=None, upcast_float32=True, data_vars='minimal',
                 coords='minimal', preprocess_func=lambda ds, **kwargs: ds):
        """Create a new NestedDictDataLoader"""
        self.file_map = file_map
        self.upcast_float32 = upcast_float32
        self.data_vars = data_vars
        self.coords = coords
        self.preprocess_func = preprocess_func

    def _generate_file_set(self, var=None, start_date=None, end_date=None,
                           domain=None, intvl_in=None, dtype_in_vert=None,
                           dtype_in_time=None, intvl_out=None):
        for name in var.names:
            try:
                return self.file_map[intvl_in][name]
            except KeyError:
                pass
        raise KeyError('Files for the var {0} cannot be found in for the '
                       'intvl_in {1} in this'
                       ' OneDirDataLoader'.format(var, intvl_in))


class GFDLDataLoader(DataLoader):
    """DataLoader for NOAA GFDL model output.

    This is an example of a domain-specific custom DataLoader, designed
    specifically for finding files output by the Geophysical Fluid Dynamics
    Laboratory's model history file post-processing tools.

    Parameters
    ----------
    template : GFDLDataLoader
        Optional argument to specify a base GFDLDataLoader to inherit
        parameters from
    data_direc : str
        Root directory of data files
    data_dur : int
        Number of years included per post-processed file
    data_start_date : datetime.datetime
        Start date of data files
    data_end_date : datetime.datetime
        End date of data files
    upcast_float32 : bool (default True)
        Whether to cast loaded DataArrays with the float32 datatype to float64
        before doing calculations
    data_vars : str (default 'minimal')
        Mode for concatenating data variables in call to ``xr.open_mfdataset``
    coords : str (default 'minimal')
        Mode for concatenating coordinate variables in call to
        ``xr.open_mfdataset``.
    preprocess_func : function (optional)
        A function to apply to every Dataset before processing in aospy.  Must
        take a Dataset and ``**kwargs`` as its two arguments.

    Examples
    --------
    Case without a template to start from.

    >>> base = GFDLDataLoader(data_direc='/archive/control/pp', data_dur=5,
    ...                       data_start_date=datetime(2000, 1, 1),
    ...                       data_end_date=datetime(2010, 12, 31))

    Case with a starting template.

    >>> data_loader = GFDLDataLoader(base, data_direc='/archive/2xCO2/pp')

    See :py:class:`aospy.data_loader.DictDataLoader` for an example of a
    possible function to pass as a ``preprocess_func``.
    """
    def __init__(self, template=None, data_direc=None, data_dur=None,
                 data_start_date=None, data_end_date=None,
                 upcast_float32=None, data_vars=None, coords=None,
                 preprocess_func=None):
        """Create a new GFDLDataLoader"""
        if template:
            _setattr_default(self, 'data_direc', data_direc,
                             getattr(template, 'data_direc'))
            _setattr_default(self, 'data_dur', data_dur,
                             getattr(template, 'data_dur'))
            _setattr_default(self, 'data_start_date', data_start_date,
                             getattr(template, 'data_start_date'))
            _setattr_default(self, 'data_end_date', data_end_date,
                             getattr(template, 'data_end_date'))
            _setattr_default(self, 'upcast_float32', upcast_float32,
                             getattr(template, 'upcast_float32'))
            _setattr_default(self, 'data_vars', data_vars,
                             getattr(template, 'data_vars'))
            _setattr_default(self, 'coords', coords,
                             getattr(template, 'coords'))
            _setattr_default(self, 'preprocess_func', preprocess_func,
                             getattr(template, 'preprocess_func'))
        else:
            self.data_direc = data_direc
            self.data_dur = data_dur
            self.data_start_date = data_start_date
            self.data_end_date = data_end_date
            _setattr_default(self, 'upcast_float32', upcast_float32, True)
            _setattr_default(self, 'data_vars', data_vars, 'minimal')
            _setattr_default(self, 'coords', coords, 'minimal')
            _setattr_default(self, 'preprocess_func', preprocess_func,
                             lambda ds, **kwargs: ds)

    @staticmethod
    def _maybe_apply_time_shift(da, time_offset=None, **DataAttrs):
        """Correct off-by-one error in GFDL instantaneous model data.

        Instantaneous data that is outputted by GFDL models is generally off by
        one timestep.  For example, a netCDF file that is supposed to
        correspond to 6 hourly data for the month of January, will have its
        last time value be in February.
        """
        if time_offset is not None:
            time = times.apply_time_offset(da[TIME_STR], **time_offset)
            da[TIME_STR] = time
        else:
            if DataAttrs['dtype_in_time'] == 'inst':
                if DataAttrs['intvl_in'].endswith('hr'):
                    offset = -1 * int(DataAttrs['intvl_in'][0])
                else:
                    offset = 0
                time = times.apply_time_offset(da[TIME_STR], hours=offset)
                da[TIME_STR] = time
        return da

    def _generate_file_set(self, var=None, start_date=None, end_date=None,
                           domain=None, intvl_in=None, dtype_in_vert=None,
                           dtype_in_time=None, intvl_out=None):
        attempted_file_sets = []
        for name in var.names:
            file_set = self._input_data_paths_gfdl(
                name, start_date, end_date, domain, intvl_in, dtype_in_vert,
                dtype_in_time, intvl_out)
            attempted_file_sets.append(file_set)
            if all([os.path.isfile(filename) for filename in file_set]):
                return file_set
        raise IOError('Files for the var {0} cannot be located '
                      'using GFDL post-processing conventions. '
                      'Attempted using the following sets of paths:\n\n'
                      '{1}'.format(var, pprint.pformat(attempted_file_sets)))

    def _input_data_paths_gfdl(self, name, start_date, end_date, domain,
                               intvl_in, dtype_in_vert, dtype_in_time,
                               intvl_out):
        dtype_lbl = dtype_in_time
        if intvl_in == 'daily':
            domain += '_daily'
        if dtype_in_vert == ETA_STR and name != 'ps':
            domain += '_level'
        if dtype_in_time == 'inst':
            domain += '_inst'
            dtype_lbl = 'ts'
        if 'monthly_from_' in dtype_in_time:
            dtype = dtype_in_time.replace('monthly_from_', '')
            dtype_lbl = dtype
        else:
            dtype = dtype_in_time
        dur_str = str(self.data_dur) + 'yr'
        if dtype_in_time == 'av':
            subdir = intvl_in + '_' + dur_str
        else:
            subdir = os.path.join(intvl_in, dur_str)
        direc = os.path.join(self.data_direc, domain, dtype_lbl, subdir)
        data_start_year = times.infer_year(self.data_start_date)
        start_year = times.infer_year(start_date)
        end_year = times.infer_year(end_date)
        files = [os.path.join(direc, io.data_name_gfdl(
                    name, domain, dtype, intvl_in, year, intvl_out,
                    data_start_year, self.data_dur))
                 for year in range(start_year, end_year + 1)]
        files = list(set(files))
        files.sort()
        return files



"""Funcionality for representing a physical variable in aospy."""
import numpy as np


class Var(object):
    """An object representing a physical quantity to be computed.

    Attributes
    ----------
    name : str
        The variable's name
    alt_names : tuple of strings
        All other names that the variable may be referred to in the input data
    names : tuple of strings
        The combination of `name` and `alt_names`
    description : str
        A description of the variable
    func : function
        The function with which to compute the variable
    variables : sequence of aospy.Var objects
        The variables passed to `func` to compute it
    units : str
        The variable's physical units
    domain : str
        The physical domain of the variable, e.g. 'atmos', 'ocean', or 'land'
    def_time, def_vert, def_lat, def_lon : bool
        Whether the variable is defined, respectively, in time, vertically, in
        latitude, and in longitude
    math_str : str
        The mathematical representation of the variable
    colormap : str
        The name of the default colormap to be used in plots of this variable
    valid_range : length-2 tuple
        The range of values outside which to flag as unphysical/erroneous

    """

    def __init__(self, name, alt_names=None, func=None, variables=None,
                 units='', plot_units='', plot_units_conv=1, domain='atmos',
                 description='', def_time=False, def_vert=False, def_lat=False,
                 def_lon=False, math_str=False, colormap='RdBu_r',
                 valid_range=None):
        """Instantiate a Var object.

        Parameters
        ----------
        name : str
            The variable's name
        alt_names : tuple of strings
            All other names that the variable might be referred to in any input
            data.  Each of these should be unique to this variable in order to
            avoid loading the wrong quantity.
        description : str
            A description of the variable
        func : function
            The function with which to compute the variable
        variables : sequence of aospy.Var objects
            The variables passed to `func` to compute it.  Order matters:
            whenever calculations are performed to generate data corresponding
            to this Var, the data corresponding to the elements of `variables`
            will be passed to `self.function` in the same order.
        units : str
            The variable's physical units
        domain : str
            The physical domain of the variable, e.g. 'atmos', 'ocean', or
            'land'.  This is only used by aospy by some types of `DataLoader`,
            including `GFDLDataLoader`.
        def_time, def_vert, def_lat, def_lon : bool
            Whether the variable is defined, respectively, in time, vertically,
            in latitude, and in longitude
        math_str : str
            The mathematical representation of the variable.  This is typically
            a raw string of LaTeX math-mode, e.g. r'$T_\mathrm{sfc}$' for
            surface temperature.
        colormap : str
            (Currently not used by aospy) The name of the default colormap to
            be used in plots of this variable.
        valid_range : length-2 tuple
            The range of values outside which to flag as unphysical/erroneous
        """  # noqa: W605
        self.name = name
        if alt_names is None:
            self.names = tuple([name])
        else:
            self.alt_names = alt_names
            self.names = tuple([name] + list(alt_names))

        if func is None:
            self.func = lambda x: x
            self.variables = None
        else:
            self.func = func
            self.variables = variables

        self.units = units

        if not description:
            if self.func.__doc__ is None:
                self.description = ''
            else:
                self.description = self.func.__doc__
        else:
            self.description = description

        self.domain = domain
        self.def_time = def_time
        self.def_vert = def_vert
        self.def_lat = def_lat
        self.def_lon = def_lon
        self.math_str = math_str
        self.colormap = colormap
        self.valid_range = valid_range

    def __str__(self):
        return 'Var instance "' + self.name + '"'

    __repr__ = __str__

    def to_plot_units(self, data, dtype_vert=False):
        """Convert the given data to plotting units."""
        if dtype_vert == 'vert_av' or not dtype_vert:
            conv_factor = self.units.plot_units_conv
        elif dtype_vert == ('vert_int'):
            conv_factor = self.units.vert_int_plot_units_conv
        else:
            raise ValueError("dtype_vert value `{0}` not recognized.  Only "
                             "bool(dtype_vert) = False, 'vert_av', and "
                             "'vert_int' supported.".format(dtype_vert))
        if isinstance(data, dict):
            return {key: val*conv_factor for key, val in data.items()}
        return data*conv_factor

    def mask_unphysical(self, data):
        """Mask data array where values are outside physically valid range."""
        if not self.valid_range:
            return data
        else:
            return np.ma.masked_outside(data, np.min(self.valid_range),
                                        np.max(self.valid_range))



"""Internal names used throughout aospy

Depending on its source, data corresponding to the same physical
variable can have different internal names. To ensure that all
variables can be easily accessed, upon loading in data ``aospy``
overwrites these names with names specified either in ``aospy.Var``
objects (in the case of variables like 'precip'), or names specified
in this module (in the case of frequently encountered coordinate variables,
like 'latitude').

The ``GRID_ATTRS`` dictionary maps the internal names for coordinates used
in ``aospy`` to lists of potential names based on all previously encountered
names for each coordinate in different datasets.  ``aospy`` uses this
dictionary to identify given physical coordinates, and rename them with
their proper internal names.
"""
from collections import OrderedDict

# Horizontal coordinates
LON_STR = 'lon'
LAT_STR = 'lat'
LON_BOUNDS_STR = 'lon_bounds'
LAT_BOUNDS_STR = 'lat_bounds'
SFC_AREA_STR = 'sfc_area'
LAND_MASK_STR = 'land_mask'

# Vertical coordinates
ETA_STR = 'sigma'
PHALF_STR = 'phalf'
PFULL_STR = 'pfull'
PLEVEL_STR = 'level'
PK_STR = 'pk'
BK_STR = 'bk'
ZSURF_STR = 'zsurf'

# Time coordinates
TIME_STR = 'time'
TIME_BOUNDS_STR = 'time_bounds'
YEAR_STR = 'year'
BOUNDS_STR = 'bounds'
AVERAGE_T1_STR = 'average_T1'
AVERAGE_T2_STR = 'average_T2'
TIME_WEIGHTS_STR = 'time_weights'
RAW_START_DATE_STR = 'raw_data_start_date'
RAW_END_DATE_STR = 'raw_data_end_date'
SUBSET_START_DATE_STR = 'subset_start_date'
SUBSET_END_DATE_STR = 'subset_end_date'
TIME_VAR_STRS = [TIME_STR, TIME_BOUNDS_STR, TIME_WEIGHTS_STR]

# All attributes associated with data's spatiotemporal grid.
GRID_ATTRS = OrderedDict(
    [(LAT_STR, ('lat', 'latitude', 'LATITUDE', 'y', 'Y', 'yto', 'XLAT')),
     (LAT_BOUNDS_STR, ('latb', 'lat_bnds', 'lat_bounds')),
     (LON_STR, ('lon', 'longitude', 'LONGITUDE', 'x', 'X', 'xto', 'XLONG')),
     (LON_BOUNDS_STR, ('lonb', 'lon_bnds', 'lon_bounds')),
     (ZSURF_STR, ('zsurf', 'HGT')),
     (SFC_AREA_STR, ('area', 'sfc_area')),
     (LAND_MASK_STR, ('land_mask', 'LANDFRAC', 'XLAND', 'land')),
     (PK_STR, ('pk',)),
     (BK_STR, ('bk',)),
     (PHALF_STR, ('phalf',)),
     (PFULL_STR, ('pfull',)),
     (PLEVEL_STR, ('level', 'lev', 'plev', 'P')),
     (TIME_STR, ('time', 'XTIME')),
     (TIME_WEIGHTS_STR, ('time_weights', 'average_DT',)),
     (TIME_BOUNDS_STR, ('time_bounds', 'time_bnds')),
     (BOUNDS_STR, ('bounds', 'bnds', 'nv', 'nbnd', 'nbnds')),
     (RAW_START_DATE_STR, ('raw_data_start_date',)),
     (RAW_END_DATE_STR, ('raw_data_end_date',))]
)

GRID_ATTRS_NO_TIMES = GRID_ATTRS.copy()
for key in TIME_VAR_STRS:
    del GRID_ATTRS_NO_TIMES[key]



"""Functionality for representing individual projects for use in aospy."""
import logging
import time


class Proj(object):
    """An object that describes a single project that will use aospy.

    This is the top-level class in the aospy hierarchy of data representations.
    It is meant to contain all of the `Model`, `Run`, and `Region` objects that
    are of relevance to a particular research project.  (Any of these may be
    used by multiple `Proj` objects.)

    The `Proj` class itself provides little functionality, but it is an
    important means of organizing a user's work across different
    projects.  In particular, the output of all calculations using
    `aospy.Calc` are saved in a directory structure whose root is that
    of the `Proj` object specified for that calculation.


    Attributes
    ----------
    name : str
        The run's name
    description : str
        A description of the run
    direc_out, tar_direc_out : str
        The paths to the root directories of, respectively, the standard and
        .tar versions of the output of aospy calculations saved to disk.
    models : dict
        A dictionary with entries of the form ``{model_obj.name: model_obj}``,
        for each of this ``Proj``'s child model objects
    default_models : dict
        The default model objects on which to perform calculations via
        `aospy.Calc` if not otherwise specified
    regions : dict
        A dictionary with entries of the form ``{regin_obj.name: region_obj}``,
        for each of this ``Proj``'s child region objects
"""

    def __init__(self, name, description=None, models=None,
                 default_models=None, regions=None, direc_out='',
                 tar_direc_out=''):
        """
        Parameters
        ----------
        name : str
            The project's name.  This should be unique from that of any other
            `Proj` objects being used.
        description : str, optional
            A description of the model.  This is not used internally by
            aospy; it is solely for the user's information.
        regions : {None, sequence of aospy.Region objects}, optional
            The desired regions over which to perform region-average
            calculations.
        models : {None, sequence of aospy.Model objects}, optional
            The child Model objects of this project.
        default_models : {None, sequence of aospy.Run objects}, optional
            The subset of this Model's runs over which to perform calculations
            by default.
        direc_out, tar_direc_out : str
            Path to the root directories of where, respectively, regular output
            and a .tar-version of the output will be saved to disk.

        Notes
        -----
        Instantiating a `Proj` object has the side-effect of setting the `proj`
        attribute of each of it's child `Model` objects to itself.

        See Also
        --------
        aospy.Model, aospy.Region, aospy.Run

        """
        logging.debug("Initializing Project instance: %s (%s)"
                      % (name, time.ctime()))
        self.name = name
        self.description = '' if description is None else description
        self.direc_out = direc_out
        self.tar_direc_out = tar_direc_out

        if models is None:
            self.models = []
        else:
            self.models = models
        if default_models is None:
            self.default_models = []
        else:
            self.default_models = default_models
        if regions is None:
            self.regions = []
        else:
            self.regions = regions

        # Set the `proj` attribute of the children models.
        for model in self.models:
            setattr(model, 'proj', self)

    def __str__(self):
        return 'Project instance "' + self.name + '"'

    __repr__ = __str__



from multiprocessing import cpu_count
from os.path import isfile
import shutil
import sys
import itertools

import distributed
import pytest

from aospy import Var, Proj
from aospy.automate import (_get_attr_by_tag, _permuted_dicts_of_specs,
                            _get_all_objs_of_type, _merge_dicts,
                            _input_func_py2_py3, AospyException,
                            _user_verify, CalcSuite, _MODELS_STR, _RUNS_STR,
                            _VARIABLES_STR, _REGIONS_STR,
                            _compute_or_skip_on_error, submit_mult_calcs,
                            _n_workers_for_local_cluster,
                            _prune_invalid_time_reductions)
from .data.objects import examples as lib
from .data.objects.examples import (
    example_proj, example_model, example_run, var_not_time_defined,
    condensation_rain, convection_rain, precip, ps, sphum, globe, sahel, bk,
    p, dp
)


@pytest.fixture
def obj_lib():
    return lib


@pytest.fixture
def all_vars():
    return [condensation_rain, convection_rain, precip, ps, sphum]


@pytest.fixture
def all_projects():
    return [example_proj]


@pytest.fixture
def all_models():
    return [example_model]


@pytest.fixture
def all_runs():
    return [example_run]


@pytest.fixture
def all_regions():
    return [globe, sahel]


@pytest.mark.parametrize(
    ('obj', 'tag', 'attr_name', 'expected'),
    [(example_proj, 'all', _MODELS_STR, [example_model]),
     (example_proj, 'default', _MODELS_STR, []),
     (example_model, 'all', _RUNS_STR, [example_run]),
     (example_model, 'default', _RUNS_STR, [])])
def test_get_attr_by_tag(obj, tag, attr_name, expected):
    actual = _get_attr_by_tag(obj, tag, attr_name)
    assert actual == expected


def test_get_attr_by_tag_invalid():
    with pytest.raises(KeyError):
        _get_attr_by_tag(example_proj, 'alll', _MODELS_STR)


@pytest.fixture
def calcsuite_specs():
    """Aux specs after being processed by CalcSuite."""
    return {
        'time_offset': [None],
        'date_range': ['default'],
        'intvl_in': ['monthly'],
        'region': [{globe, sahel}],
        'dtype_out_time': [['av', 'reg.av']],
        'dtype_in_vert': [False],
        'dtype_in_time': ['ts'],
        'var': [condensation_rain, convection_rain],
        'intvl_out': ['ann'],
        'dtype_out_vert': [None]
    }


def test_permuted_dict_of_specs(calcsuite_specs):
    actual = _permuted_dicts_of_specs(calcsuite_specs)
    expected = [
        {'time_offset': None,
         'date_range': 'default',
         'intvl_in': 'monthly',
         'region': {globe, sahel},
         'dtype_out_time': ['av', 'reg.av'],
         'dtype_in_vert': False,
         'dtype_in_time': 'ts',
         'var': condensation_rain,
         'intvl_out': 'ann',
         'dtype_out_vert': None},
        {'time_offset': None,
         'date_range': 'default',
         'intvl_in': 'monthly',
         'region': {globe, sahel},
         'dtype_out_time': ['av', 'reg.av'],
         'dtype_in_vert': False,
         'dtype_in_time': 'ts',
         'var': convection_rain,
         'intvl_out': 'ann',
         'dtype_out_vert': None}
    ]
    assert actual == expected


def test_merge_dicts():
    # no conflicts
    dict1 = dict(a=1)
    dict2 = {'b': 3, 43: False}
    dict3 = dict(c=['abc'])
    expected = {'a': 1, 'b': 3, 'c': ['abc'], 43: False}
    assert expected == _merge_dicts(dict1, dict2, dict3)

    # conflicts
    dict4 = dict(c=None)
    expected = {'a': 1, 'b': 3, 'c': None, 43: False}
    assert expected == _merge_dicts(dict1, dict2, dict3, dict4)


def test_input_func_py2_py3():
    result = _input_func_py2_py3()
    if sys.version.startswith('3'):
        import builtins
        assert result is builtins.input
    elif sys.version.startswith('2'):
        assert result is raw_input  # noqa: F821


def test_user_verify():
    with pytest.raises(AospyException):
        _user_verify(lambda x: 'no')
    _user_verify(lambda x: 'YES')


@pytest.mark.parametrize(
    ('type_', 'expected'),
    [(Var, [var_not_time_defined, condensation_rain, convection_rain,
            precip, ps, sphum, bk, p, dp]),
     (Proj, [example_proj])])
def test_get_all_objs_of_type(obj_lib, type_, expected):
    actual = _get_all_objs_of_type(type_, obj_lib)
    assert set(expected) == set(actual)


@pytest.fixture
def calcsuite_init_specs():
    return dict(
        library=lib,
        projects=[example_proj],
        models=[example_model],
        runs=[example_run],
        variables=[condensation_rain, convection_rain],
        regions='all',
        date_ranges='default',
        output_time_intervals=['ann'],
        output_time_regional_reductions=['av', 'reg.av'],
        output_vertical_reductions=[None],
        input_time_intervals=['monthly'],
        input_time_datatypes=['ts'],
        input_time_offsets=[None],
        input_vertical_datatypes=[False],
    )


@pytest.fixture
def calcsuite_init_specs_single_calc(calcsuite_init_specs):
    specs = calcsuite_init_specs.copy()
    specs['variables'] = [condensation_rain]
    specs['regions'] = [None]
    specs['output_time_regional_reductions'] = ['av']
    yield specs
    # Teardown procedure
    for direc in [example_proj.direc_out, example_proj.tar_direc_out]:
        shutil.rmtree(direc, ignore_errors=True)


@pytest.fixture
def calcsuite_init_specs_two_calcs(calcsuite_init_specs):
    specs = calcsuite_init_specs.copy()
    specs['variables'] = [condensation_rain, convection_rain]
    specs['regions'] = [None]
    specs['output_time_regional_reductions'] = ['av']
    yield specs
    # Teardown procedure
    for direc in [example_proj.direc_out, example_proj.tar_direc_out]:
        shutil.rmtree(direc, ignore_errors=True)


@pytest.fixture
def calc(calcsuite_init_specs_single_calc):
    return CalcSuite(calcsuite_init_specs_single_calc).create_calcs()[0]


def test_compute_or_skip_on_error(calc, caplog):
    result = _compute_or_skip_on_error(calc, dict(write_to_tar=False))
    assert result is calc

    calc.start_date = 'dummy'
    result = _compute_or_skip_on_error(calc, dict(write_to_tar=False))
    log_record = caplog.record_tuples[-1][-1]
    assert log_record.startswith("Skipping aospy calculation")
    assert result is None


@pytest.fixture
def external_client():
    # Explicitly specify we want only 4 workers so that when running on
    # Travis we don't request too many.
    cluster = distributed.LocalCluster(n_workers=4)
    client = distributed.Client(cluster)
    yield client
    client.close()
    cluster.close()


def assert_calc_files_exist(calcs, write_to_tar, dtypes_out_time):
    """Check that expected calcs were written to files"""
    for calc in calcs:
        for dtype_out_time in dtypes_out_time:
            assert isfile(calc.path_out[dtype_out_time])
            if write_to_tar:
                assert isfile(calc.path_tar_out)
            else:
                assert not isfile(calc.path_tar_out)


@pytest.mark.skipif(sys.version.startswith('2'),
                    reason='https://github.com/spencerahill/aospy/issues/259')
@pytest.mark.parametrize(
    ('exec_options'),
    [dict(parallelize=True, write_to_tar=False),
     dict(parallelize=True, write_to_tar=True)])
def test_submit_mult_calcs_external_client(calcsuite_init_specs_single_calc,
                                           external_client, exec_options):
    exec_options.update(client=external_client)
    calcs = submit_mult_calcs(calcsuite_init_specs_single_calc, exec_options)
    write_to_tar = exec_options.pop('write_to_tar', True)
    assert_calc_files_exist(
        calcs, write_to_tar,
        calcsuite_init_specs_single_calc['output_time_regional_reductions'])


@pytest.mark.skipif(sys.version.startswith('2'),
                    reason='https://github.com/spencerahill/aospy/issues/259')
@pytest.mark.parametrize(
    ('exec_options'),
    [dict(parallelize=False, write_to_tar=False),
     dict(parallelize=True, write_to_tar=False),
     dict(parallelize=False, write_to_tar=True),
     dict(parallelize=True, write_to_tar=True),
     None])
def test_submit_mult_calcs(calcsuite_init_specs_single_calc, exec_options):
    calcs = submit_mult_calcs(calcsuite_init_specs_single_calc, exec_options)
    if exec_options is None:
        write_to_tar = True
    else:
        write_to_tar = exec_options.pop('write_to_tar', True)
    assert_calc_files_exist(
        calcs, write_to_tar,
        calcsuite_init_specs_single_calc['output_time_regional_reductions'])


def test_submit_mult_calcs_no_calcs(calcsuite_init_specs):
    specs = calcsuite_init_specs.copy()
    specs['input_vertical_datatypes'] = []
    with pytest.raises(AospyException):
        submit_mult_calcs(specs)


@pytest.mark.skipif(sys.version.startswith('2'),
                    reason='https://github.com/spencerahill/aospy/issues/259')
@pytest.mark.parametrize(
    ('exec_options'),
    [dict(parallelize=True, write_to_tar=False),
     dict(parallelize=True, write_to_tar=True)])
def test_submit_two_calcs_external_client(calcsuite_init_specs_two_calcs,
                                          external_client, exec_options):
    exec_options.update(client=external_client)
    calcs = submit_mult_calcs(calcsuite_init_specs_two_calcs, exec_options)
    write_to_tar = exec_options.pop('write_to_tar', True)
    assert_calc_files_exist(
        calcs, write_to_tar,
        calcsuite_init_specs_two_calcs['output_time_regional_reductions'])


@pytest.mark.skipif(sys.version.startswith('2'),
                    reason='https://github.com/spencerahill/aospy/issues/259')
@pytest.mark.parametrize(
    ('exec_options'),
    [dict(parallelize=False, write_to_tar=False),
     dict(parallelize=True, write_to_tar=False),
     dict(parallelize=False, write_to_tar=True),
     dict(parallelize=True, write_to_tar=True),
     None])
def test_submit_two_calcs(calcsuite_init_specs_two_calcs, exec_options):
    calcs = submit_mult_calcs(calcsuite_init_specs_two_calcs, exec_options)
    if exec_options is None:
        write_to_tar = True
    else:
        write_to_tar = exec_options.pop('write_to_tar', True)
    assert_calc_files_exist(
        calcs, write_to_tar,
        calcsuite_init_specs_two_calcs['output_time_regional_reductions'])


def test_n_workers_for_local_cluster(calcsuite_init_specs_two_calcs):
    calcs = CalcSuite(calcsuite_init_specs_two_calcs).create_calcs()
    expected = min(cpu_count(), len(calcs))
    result = _n_workers_for_local_cluster(calcs)
    assert result == expected


@pytest.fixture
def calc_suite(calcsuite_init_specs):
    return CalcSuite(calcsuite_init_specs)


class TestCalcSuite(object):
    def test_init(self, calc_suite, calcsuite_init_specs, obj_lib):
        assert calc_suite._specs_in == calcsuite_init_specs
        assert calc_suite._obj_lib == obj_lib

    def test_permute_core_specs(self, calc_suite):
        expected = [dict(proj=example_proj, model=example_model,
                         run=example_run)]
        actual = calc_suite._permute_core_specs()
        assert expected == actual
        # TODO: cases w/ multiple projs and/or models and/or runs, with
        #       different default children for each

    def test_get_regions(self, calc_suite, all_regions):
        assert calc_suite._get_regions()[0] == set(all_regions)
        # TODO: case w/ not all regions
        # TODO: case w/ Region objects in 'regions' sub-module

    def test_get_variables(self, calc_suite, all_vars):
        assert not hasattr(calc_suite, 'variables')
        assert calc_suite._get_variables() == {condensation_rain,
                                               convection_rain}
        # TODO: case w/ Var objects in 'variables' sub-module
        # TODO: case w/ 'all'

    def test_get_aux_specs(self, calc_suite, all_regions):
        spec_names = [name for name in calc_suite._AUX_SPEC_NAMES
                      if name not in [_VARIABLES_STR, _REGIONS_STR]]
        expected = {name: calc_suite._specs_in[name] for name in spec_names}
        expected[_VARIABLES_STR] = {condensation_rain, convection_rain}
        expected[_REGIONS_STR] = [{globe, sahel}]
        expected['date_ranges'] = ['default']
        expected['output_time_regional_reductions'] = [['av', 'reg.av']]
        actual = calc_suite._get_aux_specs()
        assert actual == expected

    def test_permute_aux_specs(self, calc_suite, calcsuite_specs):
        expected = _permuted_dicts_of_specs(calcsuite_specs)
        actual = calc_suite._permute_aux_specs()
        assert len(actual) == len(expected)
        for act in actual:
            assert act in expected


@pytest.mark.parametrize('var', [var_not_time_defined, condensation_rain])
def test_prune_invalid_time_reductions(var):
    time_options = ['av', 'std', 'ts', 'reg.av', 'reg.std', 'reg.ts']
    spec = {
        'var': var,
        'dtype_out_time': None
    }
    assert _prune_invalid_time_reductions(spec) is None
    for i in range(1, len(time_options) + 1):
        for time_option in list(itertools.permutations(time_options, i)):
            spec['dtype_out_time'] = time_option
            if spec['var'].def_time:
                assert _prune_invalid_time_reductions(spec) == time_option
            else:
                assert _prune_invalid_time_reductions(spec) == []



#!/usr/bin/env python
"""Test suite for aospy.run module."""
import sys
import unittest

import cftime
import numpy as np

from aospy.run import Run
from aospy.data_loader import DictDataLoader, GFDLDataLoader


class RunTestCase(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass


class TestRun(RunTestCase):
    def test_init_dates_valid_input(self):
        for attr in ['default_start_date', 'default_end_date']:
            for date in [None, np.datetime64('2000-01-01')]:
                run_ = Run(**{attr: date})
                self.assertEqual(date, getattr(run_, attr))

    def test_init_dates_invalid_input(self):
        for attr in ['default_start_date', 'default_end_date']:
            for date in [1985, False]:
                with self.assertRaises(TypeError):
                    Run(**{attr: date})

    def test_init_default_dates(self):
        gdl = GFDLDataLoader(data_start_date=cftime.DatetimeNoLeap(1, 1, 1),
                             data_end_date=cftime.DatetimeNoLeap(1, 12, 31))
        run_ = Run(data_loader=gdl)
        self.assertEqual(run_.default_start_date,
                         cftime.DatetimeNoLeap(1, 1, 1))
        self.assertEqual(run_.default_end_date,
                         cftime.DatetimeNoLeap(1, 12, 31))

        ddl = DictDataLoader({'monthly': '/a/'})
        run_ = Run(data_loader=ddl)
        self.assertEqual(run_.default_start_date, None)
        self.assertEqual(run_.default_end_date, None)


if __name__ == '__main__':
    sys.exit(unittest.main())



"""aospy: management, and analysis of gridded climate data."""
from .__config__ import user_path
from . import _constants
from . import utils
from . import var
from .var import Var
from . import region
from .region import Region
from . import run
from .run import Run
from . import model
from .model import Model
from . import proj
from .proj import Proj
from . import calc
from .calc import Calc
from .automate import submit_mult_calcs
from . import examples

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions

__all__ = ['user_path', '_constants', 'utils', 'var', 'Var', 'region',
           'Region', 'run', 'Run', 'model', 'Model', 'proj', 'Proj',
           'calc', 'Calc', 'submit_mult_calcs', 'examples']



"""Sample aospy object library using the included example data."""
from datetime import datetime
import os

import aospy
from aospy import Model, Proj, Region, Run, Var
from aospy.data_loader import DictDataLoader
from aospy.internal_names import LAND_MASK_STR, LON_STR


rootdir = os.path.join(aospy.__path__[0], 'test', 'data', 'netcdf')


_file_map = {'monthly': os.path.join(rootdir,
                                     '000[4-6]0101.precip_monthly.nc')}
example_run = Run(
    name='example_run',
    description=(
        'Control simulation of the idealized moist model'
    ),
    default_start_date=datetime(4, 1, 1),
    default_end_date=datetime(6, 12, 31),
    data_loader=DictDataLoader(_file_map)
)


example_model = Model(
    name='example_model',
    grid_file_paths=(os.path.join(rootdir, '00040101.precip_monthly.nc'),
                     os.path.join(rootdir, 'im.landmask.nc')),
    runs=[example_run],
    grid_attrs={LAND_MASK_STR: 'custom_land_mask', LON_STR: 'custom_lon'}
)


def total_precip(precip_largescale, precip_convective):
    """Sum of convective and large-scale precipitation.

    Parameters
    ----------
    precip_largescale, precip_convective : xarray.DataArrays
        Precipitation from grid-scale condensation and from convective
        parameterization, respectively.

    Returns
    -------
    xarray.DataArray
    """
    return precip_largescale + precip_convective


def conv_precip_frac(precip_largescale, precip_convective):
    """Fraction of total precip that is from convection parameterization.

    Parameters
    ----------
    precip_largescale, precip_convective : xarray.DataArrays
        Precipitation from grid-scale condensation and from convective
        parameterization, respectively.

    Returns
    -------
    xarray.DataArray
    """
    total = total_precip(precip_largescale, precip_convective)
    # Mask using xarray's `where` method to prevent divide-by-zero.
    return precip_convective / total.where(total)


precip_largescale = Var(
    name='precip_largescale',
    alt_names=('condensation_rain',),
    def_time=True,
    description='Precipitation generated via grid-scale condensation',
)


precip_convective = Var(
    name='precip_convective',
    alt_names=('convection_rain',),
    def_time=True,
    description='Precipitation generated by convective parameterization',
)


precip_total = Var(
    name='precip_total',
    def_time=True,
    func=total_precip,
    variables=(precip_largescale, precip_convective),
)


precip_conv_frac = Var(
   name='precip_conv_frac',
   def_time=True,
   func=conv_precip_frac,
   variables=(precip_largescale, precip_convective),
)


globe = Region(
    name='globe',
    description='Entire globe',
    west_bound=0,
    east_bound=360,
    south_bound=-90,
    north_bound=90,
    do_land_mask=False
)


tropics = Region(
    name='tropics',
    description='Tropics, defined as 30S-30N',
    west_bound=0,
    east_bound=360,
    south_bound=-30,
    north_bound=30,
    do_land_mask=False
)


example_proj = Proj(
    'example_proj',
    direc_out='example-output',
    tar_direc_out='example-tar-output',
    models=[example_model],
    regions=(globe, tropics)
)


if __name__ == '__main__':
    pass



"""Utility functions for data input and output."""
import logging
import subprocess

import numpy as np


def data_in_label(intvl_in, dtype_in_time, dtype_in_vert=False):
    """Create string label specifying the input data of a calculation."""
    intvl_lbl = intvl_in
    time_lbl = dtype_in_time
    lbl = '_'.join(['from', intvl_lbl, time_lbl]).replace('__', '_')
    vert_lbl = dtype_in_vert if dtype_in_vert else False
    if vert_lbl:
        lbl = '_'.join([lbl, vert_lbl]).replace('__', '_')
    return lbl


def data_out_label(time_intvl, dtype_time, dtype_vert=False):
    intvl_lbl = time_label(time_intvl, return_val=False)
    time_lbl = dtype_time
    lbl = '.'.join([intvl_lbl, time_lbl]).replace('..', '.')
    vert_lbl = dtype_vert if dtype_vert else False
    if vert_lbl:
        lbl = '.'.join([lbl, vert_lbl]).replace('..', '.')
    return lbl


def yr_label(yr_range):
    """Create label of start and end years for aospy data I/O."""
    assert yr_range is not None, "yr_range is None"
    if yr_range[0] == yr_range[1]:
        return '{:04d}'.format(yr_range[0])
    else:
        return '{:04d}-{:04d}'.format(*yr_range)


def time_label(intvl, return_val=True):
    """Create time interval label for aospy data I/O."""
    # Monthly labels are 2 digit integers: '01' for jan, '02' for feb, etc.
    if type(intvl) in [list, tuple, np.ndarray] and len(intvl) == 1:
        label = '{:02}'.format(intvl[0])
        value = np.array(intvl)
    elif type(intvl) == int and intvl in range(1, 13):
        label = '{:02}'.format(intvl)
        value = np.array([intvl])
    # Seasonal and annual time labels are short strings.
    else:
        labels = {'jfm': (1, 2, 3),
                  'fma': (2, 3, 4),
                  'mam': (3, 4, 5),
                  'amj': (4, 5, 6),
                  'mjj': (5, 6, 7),
                  'jja': (6,  7,  8),
                  'jas': (7, 8, 9),
                  'aso': (8, 9, 10),
                  'son': (9, 10, 11),
                  'ond': (10, 11, 12),
                  'ndj': (11, 12, 1),
                  'djf': (1, 2, 12),
                  'jjas': (6, 7, 8, 9),
                  'djfm': (12, 1, 2, 3),
                  'ann': range(1, 13)}
        for lbl, vals in labels.items():
            if intvl == lbl or set(intvl) == set(vals):
                label = lbl
                value = np.array(vals)
                break
    if return_val:
        return label, value
    else:
        return label


def data_name_gfdl(name, domain, data_type, intvl_type, data_yr,
                   intvl, data_in_start_yr, data_in_dur):
    """Determine the filename of GFDL model data output."""
    # Determine starting year of netCDF file to be accessed.
    extra_yrs = (data_yr - data_in_start_yr) % data_in_dur
    data_in_yr = data_yr - extra_yrs
    # Determine file name. Two cases: time series (ts) or time-averaged (av).
    if data_type in ('ts', 'inst'):
        if intvl_type == 'annual':
            if data_in_dur == 1:
                filename = '.'.join([domain, '{:04d}'.format(data_in_yr),
                                     name, 'nc'])
            else:
                filename = '.'.join([domain, '{:04d}-{:04d}'.format(
                    data_in_yr, data_in_yr + data_in_dur - 1
                ), name, 'nc'])
        elif intvl_type == 'monthly':
            filename = (domain + '.{:04d}'.format(data_in_yr) + '01-' +
                        '{:04d}'.format(int(data_in_yr+data_in_dur-1)) +
                        '12.' + name + '.nc')
        elif intvl_type == 'daily':
            filename = (domain + '.{:04d}'.format(data_in_yr) + '0101-' +
                        '{:04d}'.format(int(data_in_yr+data_in_dur-1)) +
                        '1231.' + name + '.nc')
        elif 'hr' in intvl_type:
            filename = '.'.join(
                [domain, '{:04d}010100-{:04d}123123'.format(
                    data_in_yr, data_in_yr + data_in_dur - 1), name, 'nc']
            )
    elif data_type == 'av':
        if intvl_type in ['annual', 'ann']:
            label = 'ann'
        elif intvl_type in ['seasonal', 'seas']:
            label = intvl.upper()
        elif intvl_type in ['monthly', 'mon']:
            label, val = time_label(intvl)
        if data_in_dur == 1:
            filename = (domain + '.{:04d}'.format(data_in_yr) +
                        '.' + label + '.nc')
        else:
            filename = (domain + '.{:04d}'.format(data_in_yr) + '-' +
                        '{:04d}'.format(int(data_in_yr+data_in_dur-1)) +
                        '.' + label + '.nc')
    elif data_type == 'av_ts':
        filename = (domain + '.{:04d}'.format(data_in_yr) + '-' +
                    '{:04d}'.format(int(data_in_yr+data_in_dur-1)) +
                    '.01-12.nc')
    return filename


def dmget(files_list):
    """Call GFDL command 'dmget' to access archived files."""
    if isinstance(files_list, str):
        files_list = [files_list]

    archive_files = []
    for f in files_list:
        if f.startswith('/archive'):
            archive_files.append(f)
    try:
        subprocess.call(['dmget'] + archive_files)
    except OSError:
        logging.debug('dmget command not found in this machine')



#!/usr/bin/env python
"""Test suite for aospy.io module."""
import sys
import unittest

import aospy.utils.io as io


class AospyIOTestCase(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass


class TestIO(AospyIOTestCase):
    def test_dmget(self):
        io.dmget(['/home/Spencer.Clark/archive/imr_skc/control/'
                  'gfdl.ncrc3-default-repro/1/history/'
                  '00010101.atmos_month.nc'])

        io.dmget('/home/Spencer.Clark/archive/imr_skc/control/'
                 'gfdl.ncrc3-default-repro/1/history/'
                 '00010101.atmos_month.nc')


if __name__ == '__main__':
    sys.exit(unittest.main())



"""Utility functions for handling times, dates, etc."""
import datetime
import logging
import re
import warnings

import cftime
import numpy as np
import pandas as pd
import xarray as xr

from ..internal_names import (
    BOUNDS_STR, RAW_END_DATE_STR, RAW_START_DATE_STR,
    SUBSET_END_DATE_STR, SUBSET_START_DATE_STR, TIME_BOUNDS_STR, TIME_STR,
    TIME_WEIGHTS_STR
)


def apply_time_offset(time, years=0, months=0, days=0, hours=0):
    """Apply a specified offset to the given time array.

    This is useful for GFDL model output of instantaneous values.  For example,
    3 hourly data postprocessed to netCDF files spanning 1 year each will
    actually have time values that are offset by 3 hours, such that the first
    value is for 1 Jan 03:00 and the last value is 1 Jan 00:00 of the
    subsequent year.  This causes problems in xarray, e.g. when trying to group
    by month.  It is resolved by manually subtracting off those three hours,
    such that the dates span from 1 Jan 00:00 to 31 Dec 21:00 as desired.

    Parameters
    ----------
    time : xarray.DataArray representing a timeseries
    years, months, days, hours : int, optional
        The number of years, months, days, and hours, respectively, to offset
        the time array by.  Positive values move the times later.

    Returns
    -------
    pandas.DatetimeIndex

    Examples
    --------
    Case of a length-1 input time array:

    >>> times = xr.DataArray(datetime.datetime(1899, 12, 31, 21))
    >>> apply_time_offset(times)
    Timestamp('1900-01-01 00:00:00')

    Case of input time array with length greater than one:

    >>> times = xr.DataArray([datetime.datetime(1899, 12, 31, 21),
    ...                       datetime.datetime(1899, 1, 31, 21)])
    >>> apply_time_offset(times) # doctest: +NORMALIZE_WHITESPACE
    DatetimeIndex(['1900-01-01', '1899-02-01'], dtype='datetime64[ns]',
                  freq=None)
    """
    return (pd.to_datetime(time.values) +
            pd.tseries.offsets.DateOffset(years=years, months=months,
                                          days=days, hours=hours))


def average_time_bounds(ds):
    """Return the average of each set of time bounds in the Dataset.

    Useful for creating a new time array to replace the Dataset's native time
    array, in the case that the latter matches either the start or end bounds.
    This can cause errors in grouping (akin to an off-by-one error) if the
    timesteps span e.g. one full month each.  Note that the Dataset's times
    must not have already undergone "CF decoding", wherein they are converted
    from floats using the 'units' attribute into datetime objects.

    Parameters
    ----------
    ds : xarray.Dataset
        A Dataset containing a time bounds array with name matching
        internal_names.TIME_BOUNDS_STR.  This time bounds array must have two
        dimensions, one of which's coordinates is the Dataset's time array, and
        the other is length-2.

    Returns
    -------
    xarray.DataArray
        The mean of the start and end times of each timestep in the original
        Dataset.

    Raises
    ------
    ValueError
        If the time bounds array doesn't match the shape specified above.

    """
    bounds = ds[TIME_BOUNDS_STR]
    new_times = bounds.mean(dim=BOUNDS_STR, keep_attrs=True)
    new_times = new_times.drop(TIME_STR).rename(TIME_STR)
    new_times[TIME_STR] = new_times
    return new_times


def monthly_mean_ts(arr):
    """Convert a sub-monthly time-series into one of monthly means.

    Also drops any months with no data in the original DataArray.

    Parameters
    ----------
    arr : xarray.DataArray
        Timeseries of sub-monthly temporal resolution data

    Returns
    -------
    xarray.DataArray
        Array resampled to comprise monthly means

    See Also
    --------
    monthly_mean_at_each_ind : Copy monthly means to each submonthly time

    """
    return arr.resample(**{TIME_STR: '1M'}).mean(TIME_STR).dropna(TIME_STR)


def monthly_mean_at_each_ind(monthly_means, sub_monthly_timeseries):
    """Copy monthly mean over each time index in that month.

    Parameters
    ----------
    monthly_means : xarray.DataArray
        array of monthly means
    sub_monthly_timeseries : xarray.DataArray
        array of a timeseries at sub-monthly time resolution

    Returns
    -------
    xarray.DataArray with eath monthly mean value from `monthly_means` repeated
    at each time within that month from `sub_monthly_timeseries`

    See Also
    --------
    monthly_mean_ts : Create timeseries of monthly mean values
    """
    time = monthly_means[TIME_STR]
    start = time.indexes[TIME_STR][0].replace(day=1, hour=0)
    end = time.indexes[TIME_STR][-1]
    new_indices = pd.DatetimeIndex(start=start, end=end, freq='MS')
    arr_new = monthly_means.reindex(time=new_indices, method='backfill')
    return arr_new.reindex_like(sub_monthly_timeseries, method='pad')


def yearly_average(arr, dt):
    """Average a sub-yearly time-series over each year.

    Resulting timeseries comprises one value for each year in which the
    original array had valid data.  Accounts for (i.e. ignores) masked values
    in original data when computing the annual averages.

    Parameters
    ----------
    arr : xarray.DataArray
        The array to be averaged
    dt : xarray.DataArray
        Array of the duration of each timestep

    Returns
    -------
    xarray.DataArray
        Has the same shape and mask as the original ``arr``, except for the
        time dimension, which is truncated to one value for each year that
        ``arr`` spanned

    """
    assert_matching_time_coord(arr, dt)
    yr_str = TIME_STR + '.year'
    # Retain original data's mask.
    dt = dt.where(np.isfinite(arr))
    return ((arr*dt).groupby(yr_str).sum(TIME_STR) /
            dt.groupby(yr_str).sum(TIME_STR))


def ensure_datetime(obj):
    """Return the object if it is a datetime-like object

    Parameters
    ----------
    obj : Object to be tested.

    Returns
    -------
    The original object if it is a datetime-like object

    Raises
    ------
    TypeError if `obj` is not datetime-like
    """
    _VALID_TYPES = (str, datetime.datetime, cftime.datetime,
                    np.datetime64)
    if isinstance(obj, _VALID_TYPES):
        return obj
    raise TypeError("datetime-like object required.  "
                    "Type given: {}".format(type(obj)))


def datetime_or_default(date, default):
    """Return a datetime-like object or a default.

    Parameters
    ----------
    date : `None` or datetime-like object or str
    default : The value to return if `date` is `None`

    Returns
    -------
    `default` if `date` is `None`, otherwise returns the result of
    `utils.times.ensure_datetime(date)`

    """
    if date is None:
        return default
    else:
        return ensure_datetime(date)


def month_indices(months):
    """Convert string labels for months to integer indices.

    Parameters
    ----------
     months : str, int
         If int, number of the desired month, where January=1, February=2,
         etc.  If str, must match either 'ann' or some subset of
         'jfmamjjasond'.  If 'ann', use all months.  Otherwise, use the
         specified months.

    Returns
    -------
    np.ndarray of integers corresponding to desired month indices

    Raises
    ------
    TypeError : If `months` is not an int or str

    See also
    --------
    _month_conditional
    """
    if not isinstance(months, (int, str)):
        raise TypeError("`months` must be of type int or str: "
                        "type(months) == {}".format(type(months)))
    if isinstance(months, int):
        return [months]
    if months.lower() == 'ann':
        return np.arange(1, 13)
    first_letter = 'jfmamjjasond' * 2
    # Python indexing starts at 0; month indices start at 1 for January.
    count = first_letter.count(months)
    if (count == 0) or (count > 2):
        message = ("The user must provide a unique pattern of consecutive "
                   "first letters of months within '{}'. The provided "
                   "string '{}' does not comply."
                   "  For individual months use integers."
                   "".format(first_letter, months))
        raise ValueError(message)
    st_ind = first_letter.find(months.lower())
    return np.arange(st_ind, st_ind + len(months)) % 12 + 1


def _month_conditional(time, months):
    """Create a conditional statement for selecting data in a DataArray.

    Parameters
    ----------
    time : xarray.DataArray
         Array of times for which to subsample for specific months.
    months : int, str, or xarray.DataArray of times
        If int or str, passed to `month_indices`
    Returns
    -------
    Array of bools specifying which months to keep

    See Also
    --------
    month_indices
    """
    if isinstance(months, (int, str)):
        months_array = month_indices(months)
    else:
        months_array = months
    cond = False
    for month in months_array:
        cond |= (time['{}.month'.format(TIME_STR)] == month)
    return cond


def extract_months(time, months):
    """Extract times within specified months of the year.

    Parameters
    ----------
    time : xarray.DataArray
         Array of times that can be represented by numpy.datetime64 objects
         (i.e. the year is between 1678 and 2262).
    months : Desired months of the year to include

    Returns
    -------
    xarray.DataArray of the desired times
    """
    inds = _month_conditional(time, months)
    return time.sel(time=inds)


def ensure_time_avg_has_cf_metadata(ds):
    """Add time interval length and bounds coordinates for time avg data.

    If the Dataset or DataArray contains time average data, enforce
    that there are coordinates that track the lower and upper bounds of
    the time intervals, and that there is a coordinate that tracks the
    amount of time per time average interval.

    CF conventions require that a quantity stored as time averages
    over time intervals must have time and time_bounds coordinates [1]_.
    aospy further requires AVERAGE_DT for time average data, for accurate
    time-weighted averages, which can be inferred from the CF-required
    time_bounds coordinate if needed.  This step should be done
    prior to decoding CF metadata with xarray to ensure proper
    computed timedeltas for different calendar types.

    .. [1] http://cfconventions.org/cf-conventions/v1.6.0/cf-conventions.html#_data_representative_of_cells

    Parameters
    ----------
    ds : Dataset or DataArray
        Input data

    Returns
    -------
    Dataset or DataArray
        Time average metadata attributes added if needed.
    """  # noqa: E501
    if TIME_WEIGHTS_STR not in ds:
        time_weights = ds[TIME_BOUNDS_STR].diff(BOUNDS_STR)
        time_weights = time_weights.rename(TIME_WEIGHTS_STR).squeeze()
        if BOUNDS_STR in time_weights.coords:
            time_weights = time_weights.drop(BOUNDS_STR)
        ds[TIME_WEIGHTS_STR] = time_weights

    raw_start_date = ds[TIME_BOUNDS_STR].isel(**{TIME_STR: 0, BOUNDS_STR: 0})
    ds[RAW_START_DATE_STR] = raw_start_date.reset_coords(drop=True)
    raw_end_date = ds[TIME_BOUNDS_STR].isel(**{TIME_STR: -1, BOUNDS_STR: 1})
    ds[RAW_END_DATE_STR] = raw_end_date.reset_coords(drop=True)

    for coord in [TIME_BOUNDS_STR, RAW_START_DATE_STR, RAW_END_DATE_STR]:
        ds[coord].attrs['units'] = ds[TIME_STR].attrs['units']
        if 'calendar' in ds[TIME_STR].attrs:
            ds[coord].attrs['calendar'] = ds[TIME_STR].attrs['calendar']

    unit_interval = ds[TIME_STR].attrs['units'].split('since')[0].strip()
    ds[TIME_WEIGHTS_STR].attrs['units'] = unit_interval
    return ds


def add_uniform_time_weights(ds):
    """Append uniform time weights to a Dataset.

    All DataArrays with a time coordinate require a time weights coordinate.
    For Datasets read in without a time bounds coordinate or explicit
    time weights built in, aospy adds uniform time weights at each point
    in the time coordinate.

    Parameters
    ----------
    ds : Dataset
        Input data

    Returns
    -------
    Dataset
    """
    time = ds[TIME_STR]
    unit_interval = time.attrs['units'].split('since')[0].strip()
    time_weights = xr.ones_like(time)
    time_weights.attrs['units'] = unit_interval
    del time_weights.attrs['calendar']
    ds[TIME_WEIGHTS_STR] = time_weights
    return ds


def _assert_has_data_for_time(da, start_date, end_date):
    """Check to make sure data is in Dataset for the given time range.

    Parameters
    ----------
    da : DataArray
         DataArray with a time variable
    start_date : datetime-like object or str
         start date
    end_date : datetime-like object or str
         end date

    Raises
    ------
    AssertionError
         If the time range is not within the time range of the DataArray

    """
    if isinstance(start_date, str) and isinstance(end_date, str):
        logging.warning(
            'When using strings to specify start and end dates, the check '
            'to determine if data exists for the full extent of the desired '
            'interval is not implemented.  Therefore it is possible that '
            'you are doing a calculation for a lesser interval than you '
            'specified.  If you would like this check to occur, use explicit '
            'datetime-like objects for bounds instead.')
        return

    if RAW_START_DATE_STR in da.coords:
        with warnings.catch_warnings(record=True):
            da_start = da[RAW_START_DATE_STR].values
            da_end = da[RAW_END_DATE_STR].values
    else:
        times = da.time.isel(**{TIME_STR: [0, -1]})
        da_start, da_end = times.values

    message = ('Data does not exist for requested time range: {0} to {1};'
               ' found data from time range: {2} to {3}.')
    # Add tolerance of one second, due to precision of cftime.datetimes
    tol = datetime.timedelta(seconds=1)
    if isinstance(da_start, np.datetime64):
        tol = np.timedelta64(tol, 'ns')
    range_exists = ((da_start - tol) <= start_date and
                    (da_end + tol) >= end_date)
    assert (range_exists), message.format(start_date, end_date,
                                          da_start, da_end)


def sel_time(da, start_date, end_date):
    """Subset a DataArray or Dataset for a given date range.

    Ensures that data are present for full extent of requested range.
    Appends start and end date of the subset to the DataArray.

    Parameters
    ----------
    da : DataArray or Dataset
        data to subset
    start_date : np.datetime64
        start of date interval
    end_date : np.datetime64
        end of date interval

    Returns
    ----------
    da : DataArray or Dataset
        subsetted data

    Raises
    ------
    AssertionError
        if data for requested range do not exist for part or all of
        requested range
    """
    _assert_has_data_for_time(da, start_date, end_date)
    da[SUBSET_START_DATE_STR] = xr.DataArray(start_date)
    da[SUBSET_END_DATE_STR] = xr.DataArray(end_date)
    return da.sel(**{TIME_STR: slice(start_date, end_date)})


def assert_matching_time_coord(arr1, arr2):
    """Check to see if two DataArrays have the same time coordinate.

    Parameters
    ----------
    arr1 : DataArray or Dataset
        First DataArray or Dataset
    arr2 : DataArray or Dataset
        Second DataArray or Dataset

    Raises
    ------
    ValueError
        If the time coordinates are not identical between the two Datasets
    """
    message = ('Time weights not indexed by the same time coordinate as'
               ' computed data.  This will lead to an improperly computed'
               ' time weighted average.  Exiting.\n'
               'arr1: {}\narr2: {}')
    if not (arr1[TIME_STR].identical(arr2[TIME_STR])):
        raise ValueError(message.format(arr1[TIME_STR], arr2[TIME_STR]))


def ensure_time_as_index(ds):
    """Ensures that time is an indexed coordinate on relevant quantites.

    Sometimes when the data we load from disk has only one timestep, the
    indexing of time-defined quantities in the resulting xarray.Dataset gets
    messed up, in that the time bounds array and data variables don't get
    indexed by time, even though they should.  Therefore, we need this helper
    function to (possibly) correct this.

    Note that this must be applied before CF-conventions are decoded; otherwise
    it casts ``np.datetime64[ns]`` as ``int`` values.

    Parameters
    ----------
    ds : Dataset
        Dataset with a time coordinate

    Returns
    -------
    Dataset

    """
    time_indexed_coords = {TIME_WEIGHTS_STR, TIME_BOUNDS_STR}
    time_indexed_vars = set(ds.data_vars).union(time_indexed_coords)
    time_indexed_vars = time_indexed_vars.intersection(ds.variables)
    for name in time_indexed_vars:
        if TIME_STR not in ds[name].indexes:
            da = ds[name].expand_dims(TIME_STR)
            da[TIME_STR] = ds[TIME_STR]
            ds[name] = da
    return ds


def infer_year(date):
    """Given a datetime-like object or string infer the year.

    Parameters
    ----------
    date : datetime-like object or str
        Input date

    Returns
    -------
    int

    Examples
    --------
    >>> infer_year('2000')
    2000
    >>> infer_year('2000-01')
    2000
    >>> infer_year('2000-01-31')
    2000
    >>> infer_year(datetime.datetime(2000, 1, 1))
    2000
    >>> infer_year(np.datetime64('2000-01-01'))
    2000
    >>> infer_year(DatetimeNoLeap(2000, 1, 1))
    2000
    >>>
    """
    if isinstance(date, str):
        # Look for a string that begins with four numbers; the first four
        # numbers found are the year.
        pattern = r'(?P<year>\d{4})'
        result = re.match(pattern, date)
        if result:
            return int(result.groupdict()['year'])
        else:
            raise ValueError('Invalid date string provided: {}'.format(date))
    elif isinstance(date, np.datetime64):
        return date.item().year
    else:
        return date.year


def maybe_convert_to_index_date_type(index, date):
    """Convert a datetime-like object to the index's date type.

    Datetime indexing in xarray can be done using either a pandas
    DatetimeIndex or a CFTimeIndex.  Both support partial-datetime string
    indexing regardless of the calendar type of the underlying data;
    therefore if a string is passed as a date, we return it unchanged.  If a
    datetime-like object is provided, it will be converted to the underlying
    date type of the index.  For a DatetimeIndex that is np.datetime64; for a
    CFTimeIndex that is an object of type cftime.datetime specific to the
    calendar used.

    Parameters
    ----------
    index : pd.Index
        Input time index
    date : datetime-like object or str
        Input datetime

    Returns
    -------
    date of the type appropriate for the time index of the Dataset
    """
    if isinstance(date, str):
        return date

    if isinstance(index, pd.DatetimeIndex):
        if isinstance(date, np.datetime64):
            return date
        else:
            return np.datetime64(str(date))
    else:
        date_type = index.date_type
        if isinstance(date, date_type):
            return date
        else:
            if isinstance(date, np.datetime64):
                # Convert to datetime.date or datetime.datetime object
                date = date.item()

            if isinstance(date, datetime.date):
                # Convert to a datetime.datetime object
                date = datetime.datetime.combine(
                    date, datetime.datetime.min.time())

            return date_type(date.year, date.month, date.day, date.hour,
                             date.minute, date.second, date.microsecond)



"""Functionality pertaining to aggregating data over geographical regions."""
from collections import namedtuple
import logging

import numpy as np

from .internal_names import (
    LAND_MASK_STR,
    LAT_STR,
    LON_STR,
    SFC_AREA_STR,
    YEAR_STR
)
from .utils.longitude import _maybe_cast_to_lon


def _get_land_mask(data, do_land_mask, land_mask_str=LAND_MASK_STR):
    if not do_land_mask:
        return 1
    try:
        land_mask = data[land_mask_str].copy()
    except AttributeError:
        # TODO: Implement aospy built-in land mask to default to.
        msg = ("No land mask found.  Using empty mask, which amounts to "
               "no land or ocean mask being applied.  Regions that use a "
               "land or ocean mask will therefore NOT be accurately "
               "computed.")
        logging.warning(msg)
        return 1
    try:
        percent_bool = land_mask.units.lower() in ('%', 'percent')
    except AttributeError:
        percent_bool = np.any(land_mask > 1)
    if percent_bool:
        land_mask *= 0.01
        logging.debug("Converting land mask from 0-100 to 0.0-1.0")
    if do_land_mask is True:
        return land_mask
    if do_land_mask == 'ocean':
        return 1. - land_mask
    if do_land_mask in ('strict_land', 'strict_ocean'):
        raise NotImplementedError
    msg = ("'do_land_mask' value of '{0}' is not one of the valid "
           "choices: [True, False, 'ocean', 'strict_land', "
           "'strict_ocean']").format(do_land_mask)
    raise ValueError(msg)


class BoundsRect(namedtuple('BoundsRect', ['west', 'east', 'south', 'north'])):
    """Bounding longitudes and latitudes of a given lat-lon rectangle."""
    def __new__(cls, west, east, south, north):
        new_west = _maybe_cast_to_lon(west, strict=True)
        new_east = _maybe_cast_to_lon(east, strict=True)
        return super(BoundsRect, cls).__new__(cls, new_west, new_east,
                                              south, north)

    def __repr__(self):
        return ("BoundsRect(west={0}, east={1}, south={2}, "
                "north={3}".format(self.west, self.east, self.south,
                                   self.north))


class Region(object):
    """Geographical region over which to perform averages and other reductions.

    Each `Proj` object includes a list of `Region` objects, which is used by
    `Calc` to determine which regions over which to perform time reductions
    over region-average quantities.

    Region boundaries are specified as either a single "rectangle" in latitude
    and longitude or the union of multiple such rectangles.  In addition, a
    land or ocean mask can be applied.

    Attributes
    ----------
    name : str
        The region's name
    description : str
        A description of the region
    mask_bounds : tuple
        The coordinates definining the lat-lon rectangle(s) that define(s)
        the region's boundaries
    do_land_mask
        Whether values occurring over land, ocean, or neither are excluded from
        the region, and whether the included points must be strictly land or
        ocean or if fractional land/ocean values are included.

    See Also
    --------
    aospy.Calc.region_calcs

    """

    def __init__(self, name='', description='', west_bound=None,
                 east_bound=None, south_bound=None, north_bound=None,
                 mask_bounds=None, do_land_mask=False):
        """Instantiate a Region object.

        Note that longitudes spanning (-180, 180), (0, 360), or any other range
        are all supported: -180 to 0, 180 to 360, etc. are interpreted as the
        western hemisphere, and 0-180, 360-540, etc. are interpreted as the
        eastern hemisphere.  This is true both of the region definition and of
        any data upon which the region mask is applied.

        E.g. suppose some of your data is defined on a -180 to 180 longitude
        grid, some of it is defined on a 0 to 360 grid, and some of it is
        defined on a -70 to 290 grid.  A single Region object will work with
        all three of these.

        Conversely, latitudes are always treated as -90 as the South Pole, 0 as
        the Equator, and 90 as the North Pole.  Latitudes larger than 90 are
        not physically meaningful.

        Parameters
        ----------
        name : str
             The region's name.  This must be unique from that of any other
             `Region` objects being used by the overlying `Proj`.
        description : str, optional
             A description of the region.  This is not used internally by
             aospy; it is solely for the user's information.
        west_bound, east_bound : { scalar, aospy.Longitude }, optional
             The western and eastern boundaries of the region.  All input
             longitudes are casted to ``aospy.Longitude`` objects, which
             essentially maps them to a 180W to 180E grid.  The region's
             longitudes always start at ``west_bound`` and move toward the east
             until reaching ``east_bound``.  This means that there are two
             distinct cases:

             - If, after this translation, ``west_bound`` is less than
               ``east_bound``, the region includes the points east of
               ``west_bound`` and west of ``east_bound``.
             - If ``west_bound`` is greater than ``east_bound``, then the
               region is treated as wrapping around the dateline, i.e. it's
               western-most point is ``east_bound``, and it includes all points
               moving east from there until ``west_bound``.

             If the region boundaries are more complicated than a single
             lat-lon rectangle, use ``mask_bounds`` instead.

        south_bound, north_bound : scalar, optional
             The southern, and northern boundaries, respectively, of the
             region.  If the region boundaries are more complicated than a
             single lat-lon rectangle, use `mask_bounds` instead.
        mask_bounds : sequence, optional
             Each element is a length-4 sequence of the format `(west_bound,
             east_bound, south_bound, north_bound)`, where each of these
             `_bound` arguments is of the form described above.
        do_land_mask, bool or str, optional
             Determines what, if any, land mask is applied in addition to the
             mask defining the region's boundaries.  Default `False`.  Must be
             one of ``False``, ``True``, 'ocean', 'strict_land', or
             'strict_ocean':

             - ``True``: apply the data's full land mask
             - ``False``: apply no mask
             - 'ocean': mask out land rather than ocean
             - 'strict_land': mask out all points that are not 100% land
             - 'strict_ocean': mask out all points that are not 100% ocean

        Examples
        --------
        Define a region spanning the entire globe:

        >>> globe = Region(name='globe', west_bound=0, east_bound=360,
        ...                south_bound=-90, north_bound=90, do_land_mask=False)

        Longitudes are handled as cyclic, so this definition could have
        equivalently used `west_bound=-180, east_bound=180` or `west_bound=200,
        east_bound=560`, or anything else that spanned 360 degrees total.

        Define a region corresponding to land in the mid-latitudes, which we'll
        define as land points within 30-60 degrees latitude in both
        hemispheres.  Because this region is the union of multiple lat-lon
        rectangles, it has to be defined using `mask_bounds`:

        >>> land_mid_lats = Region(name='land_midlat', do_land_mask=True,
        ...                        mask_bounds=[(-180, 180, 30, 60),
        ...                                     (-180, 180, -30, -60)])

        Define a region spanning the southern Tropical Atlantic ocean, which
        we'll take to be all ocean points between 60W and 30E and between the
        Equator and 30S:

        >>> atl_south_trop = Region(name='atl_sh_trop', west_bound=-60,
        ...                         east_bound=30, south_bound=-30,
        ...                         north_bound=0, do_land_mask='ocean')

        Define the "opposite" region, i.e. all ocean points in the southern
        Tropics *outside* of the Atlantic.  We simply swap ``west_bound`` and
        ``east_bound`` of the previous example:

        >>> non_atl_south_trop = Region(name='non_atl_sh_trop', west_bound=30,
        ...                             east_bound=-60, south_bound=-30,
        ...                             north_bound=0, do_land_mask='ocean')

        """
        self.name = name
        self.description = description
        self.do_land_mask = do_land_mask

        if mask_bounds is None:
            self.mask_bounds = tuple([BoundsRect(west_bound, east_bound,
                                                 south_bound, north_bound)])
        else:
            bounds = []
            for rect_bounds in mask_bounds:
                if len(rect_bounds) != 4:
                    raise ValueError("Each element of `mask_bounds` must be a "
                                     "length-4 array with values (west, east, "
                                     "south, north).  Value given: "
                                     "{}".format(rect_bounds))
                else:
                    bounds.append(BoundsRect(*rect_bounds))
            self.mask_bounds = tuple(bounds)

    def __str__(self):
        return 'Geographical region "' + self.name + '"'

    def _make_mask(self, data, lon_str=LON_STR, lat_str=LAT_STR):
        """Construct the mask that defines a region on a given data's grid."""
        mask = False
        for west, east, south, north in self.mask_bounds:
            if west < east:
                mask_lon = (data[lon_str] > west) & (data[lon_str] < east)
            else:
                mask_lon = (data[lon_str] < west) | (data[lon_str] > east)
            mask_lat = (data[lat_str] > south) & (data[lat_str] < north)
            mask |= mask_lon & mask_lat
        return mask

    def mask_var(self, data, lon_cyclic=True, lon_str=LON_STR,
                 lat_str=LAT_STR):
        """Mask the given data outside this region.

        Parameters
        ----------
        data : xarray.DataArray
            The array to be regionally masked.
        lon_cyclic : bool, optional (default True)
            Whether or not the longitudes of ``data`` span the whole globe,
            meaning that they should be wrapped around as necessary to cover
            the Region's full width.
        lon_str, lat_str : str, optional
            The names of the longitude and latitude dimensions, respectively,
            in the data to be masked.  Defaults are
            ``aospy.internal_names.LON_STR`` and
            ``aospy.internal_names.LON_STR``, respectively.

        Returns
        -------
        xarray.DataArray
            The original array with points outside of the region masked.

        """
        # TODO: is this still necessary?
        if not lon_cyclic:
            if self.west_bound > self.east_bound:
                raise ValueError("Longitudes of data to be masked are "
                                 "specified as non-cyclic, but Region's "
                                 "definition requires wraparound longitudes.")
        masked = data.where(self._make_mask(data, lon_str=lon_str,
                                            lat_str=lat_str))
        return masked

    def ts(self, data, lon_cyclic=True, lon_str=LON_STR, lat_str=LAT_STR,
           land_mask_str=LAND_MASK_STR, sfc_area_str=SFC_AREA_STR):
        """Create yearly time-series of region-averaged data.

        Parameters
        ----------
        data : xarray.DataArray
            The array to create the regional timeseries of
        lon_cyclic : { None, True, False }, optional (default True)
            Whether or not the longitudes of ``data`` span the whole globe,
            meaning that they should be wrapped around as necessary to cover
            the Region's full width.
        lat_str, lon_str, land_mask_str, sfc_area_str : str, optional
            The name of the latitude, longitude, land mask, and surface area
            coordinates, respectively, in ``data``.  Defaults are the
            corresponding values in ``aospy.internal_names``.

        Returns
        -------
        xarray.DataArray
            The timeseries of values averaged within the region and within each
            year, one value per year.

        """
        data_masked = self.mask_var(data, lon_cyclic=lon_cyclic,
                                    lon_str=lon_str, lat_str=lat_str)
        sfc_area = data[sfc_area_str]
        sfc_area_masked = self.mask_var(sfc_area, lon_cyclic=lon_cyclic,
                                        lon_str=lon_str, lat_str=lat_str)
        land_mask = _get_land_mask(data, self.do_land_mask,
                                   land_mask_str=land_mask_str)
        weights = sfc_area_masked * land_mask
        # Mask weights where data values are initially invalid in addition
        # to applying the region mask.
        weights = weights.where(np.isfinite(data))
        weights_reg_sum = weights.sum(lon_str).sum(lat_str)
        data_reg_sum = (data_masked * sfc_area_masked *
                        land_mask).sum(lat_str).sum(lon_str)
        return data_reg_sum / weights_reg_sum

    def av(self, data, lon_str=LON_STR, lat_str=LAT_STR,
           land_mask_str=LAND_MASK_STR, sfc_area_str=SFC_AREA_STR):
        """Time-average of region-averaged data.

        Parameters
        ----------
        data : xarray.DataArray
            The array to compute the regional time-average of
        lat_str, lon_str, land_mask_str, sfc_area_str : str, optional
            The name of the latitude, longitude, land mask, and surface area
            coordinates, respectively, in ``data``.  Defaults are the
            corresponding values in ``aospy.internal_names``.

        Returns
        -------
        xarray.DataArray
            The region-averaged and time-averaged data.

        """
        ts = self.ts(data, lon_str=lon_str, lat_str=lat_str,
                     land_mask_str=land_mask_str, sfc_area_str=sfc_area_str)
        if YEAR_STR not in ts.coords:
            return ts
        else:
            return ts.mean(YEAR_STR)

    def std(self, data, lon_str=LON_STR, lat_str=LAT_STR,
            land_mask_str=LAND_MASK_STR, sfc_area_str=SFC_AREA_STR):
        """Temporal standard deviation of region-averaged data.

        Parameters
        ----------
        data : xarray.DataArray
            The array to compute the regional time-average of
        lat_str, lon_str, land_mask_str, sfc_area_str : str, optional
            The name of the latitude, longitude, land mask, and surface area
            coordinates, respectively, in ``data``.  Defaults are the
            corresponding values in ``aospy.internal_names``.

        Returns
        -------
        xarray.DataArray
            The temporal standard deviation of the region-averaged data

        """
        ts = self.ts(data, lon_str=lon_str, lat_str=lat_str,
                     land_mask_str=land_mask_str, sfc_area_str=sfc_area_str)
        if YEAR_STR not in ts.coords:
            return ts
        else:
            return ts.std(YEAR_STR)
